#!/usr/bin/env python

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import sys

TYPE_ERRORS_IN_PRETTYPFA = True
def ts(avroType):
    """Create a human-readable type string of a type.

    Note: if titus.util.TYPE_ERRORS_IN_PRETTYPFA is ``True``, the type will be printed in PrettyPFA notation; otherwise, as a raw Avro schema.

    :type avroType: titus.datatype.AvroType
    :param avroType: type to print out
    :rtype: string
    :return: string representation of the type.
    """

    if TYPE_ERRORS_IN_PRETTYPFA:
        pretty = avscToPretty(avroType.jsonNode(set()))
        if pretty.count("\n") > 0:
            pretty = "\n    " + pretty.replace("\n", "\n    ") + "\n"
        return pretty
    else:
        return repr(avroType)

uniqueEngineNameCounter = 0
def uniqueEngineName():
    """Provide an engine name, incrementing titus.util.uniqueEngineNameCounter to ensure uniqueness of values supplied by this function."""
    sys.modules["titus.util"].uniqueEngineNameCounter += 1
    return "Engine_{0}".format(sys.modules["titus.util"].uniqueEngineNameCounter)

uniqueRecordNameCounter = 0
def uniqueRecordName():
    """Provide a record type name, incrementing titus.util.uniqueRecordNameCounter to ensure uniqueness of values supplied by this function."""
    sys.modules["titus.util"].uniqueRecordNameCounter += 1
    return "Record_{0}".format(sys.modules["titus.util"].uniqueRecordNameCounter)

uniqueEnumNameCounter = 0
def uniqueEnumName():
    """Provide an enum type name, incrementing titus.util.uniqueEnumNameCounter to ensure uniqueness of values supplied by this function."""
    sys.modules["titus.util"].uniqueEnumNameCounter += 1
    return "Enum_{0}".format(sys.modules["titus.util"].uniqueEnumNameCounter)

uniqueFixedNameCounter = 0
def uniqueFixedName():
    """Provide a fixed type name, incrementing titus.util.uniqueFixedNameCounter to ensure uniqueness of values supplied by this function."""
    sys.modules["titus.util"].uniqueFixedNameCounter += 1
    return "Fixed_{0}".format(sys.modules["titus.util"].uniqueFixedNameCounter)

def pos(dot, at):
    """Create a human-readable position of a PFA element in a JSON file."""
    if at is None or at == "":
        return "at {0}".format("" if (dot == "") else "PFA field \"" + dot + "\"")
    else:
        return "at {0}{1}".format(at, "" if (dot == "") else " (PFA field \"" + dot + "\")")

def flatten(x):
    """General-purpose list-of-lists flattening.

    :type x: list of lists of something
    :param x: list of lists to flatten
    :rtype: list of that something
    :return: flattened list
    """
    return [item for sublist in x for item in sublist]

def div(numer, denom):
    """PFA division behavior: inf and -inf for a finite number divided by zero and nan for 0/0.

    Also ensures that return type of any two numbers (including integers) is floating point.

    :type numer: number
    :param numer: the numerator
    :type denom: number
    :param denom: the denominator
    :rtype: float
    :return: fraction, including non-numeric possibilities
    """

    try:
        return numer / float(denom)
    except ZeroDivisionError:
        if numer > 0.0:
            return float("inf")
        elif numer < 0.0:
            return float("-inf")
        else:
            return float("nan")

class DynamicScope(object):
    """Keep track of a variable scope dynamically, for the sake of simulating lexical scope in a PFA document (ironically)."""

    def __init__(self, parent):
        """:type parent: titus.util.DynamicScope or ``None``
        :param parent: scope to use for resolving variables not found in this scope
        """
        self.parent = parent
        self.symbols = dict()

    def get(self, symbol):
        """Look up a variable, returning its value.

        :type symbol: string
        :param symbol: variable name to look up
        :rtype: anything
        :return: variable value
        """
        if symbol in self.symbols:
            return self.symbols[symbol]
        elif self.parent is not None:
            return self.parent.get(symbol)
        else:
            raise RuntimeError()

    def let(self, nameExpr):
        """Create a new variable.

        :type nameExpr: dict from new variable names to their initial values
        :param nameExpr: variables and values to assign
        """
        for symbol, init in list(nameExpr.items()):
            self.symbols[symbol] = init

    def set(self, nameExpr):
        """Reassign variables.

        :type nameExpr: dict from new variable names to their new values
        :param nameExpr: variables and values to assign
        """
        for symbol, value in list(nameExpr.items()):
            if symbol in self.symbols:
                self.symbols[symbol] = value
            elif self.parent is not None:
                self.parent.set(nameExpr)
            else:
                raise RuntimeError()

def callfcn(state, scope, fcn, args):
    """Helper function for calling function callbacks.

    Used in library functions that have been given functions as arguments.

    :type state: titus.genpy.ExecutionState
    :param state: execution state
    :type scope: titus.util.DynamicScope
    :param scope: dynamic scope object
    :type fcn: callable
    :param fcn: function to call (must be a PFA user-defined function or library function)
    :type args: list of values
    :param args: arguments to pass to the function (other than the ``state`` and ``scope``)
    :rtype: anything
    :return: whatever ``fcn`` returns
    """

    if hasattr(fcn, "paramNames"):
        callScope = DynamicScope(scope)
        if isinstance(args, (tuple, list)):
            args = dict(list(zip(fcn.paramNames, args)))
        callScope.let(args)
        return fcn(state, callScope)
    else:
        return fcn(state, scope, pos, None, *args)

def negativeIndex(length, index):
    """Helper function for interpreting negative array indexes as counting from the end of the array (just like Python).

    :type length: non-negative integer
    :param length: length of the array in question
    :type index: integer
    :param index: index to interpret
    :rtype: non-negative integer
    :return: if negative, count from the end of the array; otherwise, pass-through
    """

    if index >= 0:
        return index
    else:
        return length + index

def startEnd(length, start, end):
    """Helper function to normalize the start and end of an array by counting from the end of the array (just like Python).

    :type length: non-negative integer
    :param length: length of the array in question
    :type start: integer
    :param start: starting index to interpret
    :type end: integer
    :param end: ending index to interpret
    :rtype: (non-negative integer, non-negative integer)
    :return: (normalized starting index, normalized ending index)
    """

    if start >= 0:
        normStart = start
    else:
        normStart = length + start

    if normStart < 0:
        normStart = 0
    if normStart > length:
        normStart = length

    if end >= 0:
        normEnd = end
    else:
        normEnd = length + end

    if normEnd < 0:
        normEnd = 0
    if normEnd > length:
        normEnd = length
    if normEnd < normStart:
        normEnd = normStart

    return normStart, normEnd

def case(clazz):
    """Decoration to make a "case class" in Python.

    When applied to a class, read the parameters of the ``__init__`` and turn them into class fields and show them in the ``__repr__`` representation.
    """

    fields = [x for x in inspect.getargspec(clazz.__init__).args[1:] if x != "pos"]

    try:
        code = clazz.__init__.__func__
    except AttributeError:
        code = clazz.__init__.__code__

    context = dict(globals())
    context[clazz.__name__] = clazz

    if "pos" in inspect.getargspec(clazz.__init__).args:
        argFields = fields + ["pos=None"]
        assignFields = fields + ["pos"]
    else:
        argFields = assignFields = fields

    newMethods = {}
    exec("""
def __init__(self, {args}):
{assign}
    self._init({args})
""".format(args=", ".join(argFields),
           assign="\n".join(["    self.{0} = {0}".format(x) for x in (assignFields)])),
         context,
         newMethods)
    
    if len(fields) == 0:
        exec("""
def __repr__(self):
    return \"{0}()\"
""".format(clazz.__name__), context, newMethods)

        exec("""
def __eq__(self, other):
    return isinstance(other, {0})
""".format(clazz.__name__), context, newMethods)

    else:
        exec("""
def __repr__(self):
    return \"{0}(\" + {1} + \")\"
""".format(clazz.__name__, "+ \", \" + ".join(["repr(self." + x + ")" for x in fields])),
             context,
             newMethods)

        exec("""
def __eq__(self, other):
    if isinstance(other, {0}):
        return {1}
    else:
        return False
""".format(clazz.__name__, " and ".join(["self.{x} == other.{x}".format(x=x) for x in fields])),
             context,
             newMethods)

    clazz._init = clazz.__init__
    clazz.__init__ = newMethods["__init__"]

    if hasattr(clazz, "toString"):
        clazz.__repr__ = clazz.toString
    else:
        clazz.__repr__ = newMethods["__repr__"]

    if hasattr(clazz, "equals"):
        clazz.__eq__ = clazz.equals
    else:
        clazz.__eq__ = newMethods["__eq__"]

    return clazz

def untagUnion(expr, unionTypes):
    """Turn a tagged union datum into a bare Python object or pass-through if not a tagged union datum.

    :type expr: anything
    :param expr: tagged union, which may be of the form ``{"tag": value}`` or just ``value``
    :type unionTypes: list of Pythonized JSON
    :param unionTypes: types allowed by this union in Pythonized JSON schemas
    :rtype: anything
    :return: ``{"tag": value}`` collapsed to ``value`` if ``tag`` names one of the ``unionTypes``; otherwise, pass-through
    """

    if isinstance(expr, dict) and len(expr) == 1:
        tag, = list(expr.keys())
        value, = list(expr.values())
        for expectedTag in unionTypes:
            if isinstance(expectedTag, dict):
                if expectedTag["type"] in ("record", "enum", "fixed"):
                    if "namespace" in expectedTag and expectedTag["namespace"].strip() != "":
                        expectedTag = expectedTag["namespace"] + "." + expectedTag["name"]
                    else:
                        expectedTag = expectedTag["name"]
                else:
                    expectedTag = expectedTag["type"]
            if tag == expectedTag:
                return value
    return expr

def avscToPretty(avsc, indent=0):
    """Turn an Avro type from Pythonized JSON into a more human-readable PrettyPFA string.

    :type avsc: Pythonized JSON
    :param avsc: Avro type schema
    :type indent: non-negative integer
    :param indent: number of spaces to indent (default 0)
    :rtype: string
    :return: PrettyPFA representation
    """

    if isinstance(avsc, str):
        return " " * indent + avsc
    elif isinstance(avsc, dict) and "type" in avsc:
        tpe = avsc["type"]

        if tpe == "function":
            return " " * indent + "fcn(" + ", ".join("arg{}: {}".format(i + 1, avscToPretty(x)) for i, x in enumerate(avsc["params"])) + " -> " + avscToPretty(avsc["ret"]) + ")"

        elif tpe == "exception":
            return "exception"

        elif tpe in ("null", "boolean", "int", "long", "float", "double", "bytes", "string"):
            return " " * indent + tpe

        elif tpe == "array":
            return " " * indent + "array(" + avscToPretty(avsc["items"], indent + 6).lstrip() + ")"

        elif tpe == "map":
            return " " * indent + "map(" + avscToPretty(avsc["values"], indent + 4).lstrip() + ")"

        elif tpe == "record":
            name = avsc["name"]
            if "namespace" in avsc and len(avsc["namespace"]) > 0:
                name = avsc["namespace"] + "." + name
            fields = []
            for f in avsc["fields"]:
                fields.append(" " * (indent + 7) + f["name"] + ": " + avscToPretty(f["type"], indent + 7 + len(f["name"]) + 2).lstrip())
            return " " * indent + "record(" + name + ",\n" + ",\n".join(fields) + ")"

        elif tpe == "fixed":
            name = avsc["name"]
            if "namespace" in avsc and len(avsc["namespace"]) > 0:
                name = avsc["namespace"] + "." + name
            return " " * indent + "fixed(" + str(avsc["size"]) + ", " + name + ")"

        elif tpe == "enum":
            name = avsc["name"]
            if "namespace" in avsc and len(avsc["namespace"]) > 0:
                name = avsc["namespace"] + "." + name
            return " " * indent + "enum([" + ", ".join(x for x in avsc["symbols"]) + "], " + name + ")"

        else:
            raise TypeError("malformed Avro schema")

    elif isinstance(avsc, (list, tuple)):
        variants = []
        for x in avsc:
            variants.append(avscToPretty(x, indent + 6))
        return " " * indent + "union(" + ",\n".join(variants).lstrip() + ")"

    else:
        raise TypeError("malformed Avro schema")

def bytesToString(value):
    """Coverts value in bytes to corresponding string representation.

    b'hello' -> 'hello'
    b'\x80abc' -> '\x80abc'

    :type value: bytes
    :param value: the value
    :rtype: string
    :return: value in string
    """

    try:
        return ''.join(map(chr, list(value)))
    except:
        raise TypeError("value is not of type bytes")

def stringToBytes(value):
    """Coverts value in string to corresponding bytes representation.

    'hello' -> b'hello'
    '\x80abc' -> b'\x80abc'

    :type value: string
    :param value: the value
    :rtype: bytes
    :return: value in bytes
    """

    try:
        return bytes(map(ord, list(value)))
    except:
        raise TypeError("value is not of type string")           



#!/usr/bin/env python

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

from titus.fcn import Fcn
from titus.fcn import LibFcn
from titus.signature import Sig
from titus.signature import Sigs
from titus.datatype import *
from titus.errors import *
from titus.lib.core import checkForOverflow
import titus.P as P

provides = {}
def provide(fcn):
    provides[fcn.name] = fcn

prefix = "m."

#################################################################### constants (0-arity side-effect free functions)

class Pi(LibFcn):
    name = prefix + "pi"
    sig = Sig([], P.Double())
    errcodeBase = 27000
    def genpy(self, paramTypes, args, pos):
        return "math.pi"
    def __call__(self, state, scope, pos, paramTypes):
        return math.pi
provide(Pi())

class E(LibFcn):
    name = prefix + "e"
    sig = Sig([], P.Double())
    errcodeBase = 27010
    def genpy(self, paramTypes, args, pos):
        return "math.e"
    def __call__(self, state, scope, pos, paramTypes):
        return math.e
provide(E())

#################################################################### basic functions (alphabetical order)

anyNumber = set([AvroInt(), AvroLong(), AvroFloat(), AvroDouble()])

class Abs(LibFcn):
    name = prefix + "abs"
    sig = Sig([{"x": P.Wildcard("A", anyNumber)}], P.Wildcard("A"))
    errcodeBase = 27020
    def __call__(self, state, scope, pos, paramTypes, x):
        return checkForOverflow(paramTypes[0], abs(x), self.errcodeBase + 0, self.errcodeBase + 1, self.name, pos)
provide(Abs())

class ACos(LibFcn):
    name = prefix + "acos"
    sig = Sig([{"x": P.Double()}], P.Double())
    errcodeBase = 27030
    def __call__(self, state, scope, pos, paramTypes, x):
        try:
            return math.acos(x)
        except ValueError:
            return float("nan")
provide(ACos())

class ASin(LibFcn):
    name = prefix + "asin"
    sig = Sig([{"x": P.Double()}], P.Double())
    errcodeBase = 27040
    def __call__(self, state, scope, pos, paramTypes, x):
        try:
            return math.asin(x)
        except ValueError:
            return float("nan")
provide(ASin())

class ATan(LibFcn):
    name = prefix + "atan"
    sig = Sig([{"x": P.Double()}], P.Double())
    errcodeBase = 27050
    def genpy(self, paramTypes, args, pos):
        return "math.atan({0})".format(*args)
    def __call__(self, state, scope, pos, paramTypes, x):
        return math.atan(x)
provide(ATan())

class ATan2(LibFcn):
    name = prefix + "atan2"
    sig = Sig([{"y": P.Double()}, {"x": P.Double()}], P.Double())
    errcodeBase = 27060
    def genpy(self, paramTypes, args, pos):
        return "math.atan2({0}, {1})".format(*args)
    def __call__(self, state, scope, pos, paramTypes, x, y):
        return math.atan2(x, y)
provide(ATan2())

class Ceil(LibFcn):
    name = prefix + "ceil"
    sig = Sig([{"x": P.Double()}], P.Double())
    errcodeBase = 27070
#    def genpy(self, paramTypes, args, pos):
#        return "math.ceil({0})".format(*args)
    def __call__(self, state, scope, pos, paramTypes, x):
        if math.isinf(x) or math.isnan(x):
            return x
        else:
            return math.ceil(x)
provide(Ceil())

class CopySign(LibFcn):
    name = prefix + "copysign"
    sig = Sig([{"mag": P.Wildcard("A", anyNumber)}, {"sign": P.Wildcard("A")}], P.Wildcard("A"))
    errcodeBase = 27080
    def __call__(self, state, scope, pos, paramTypes, mag, sign):
        return abs(mag) * (-1 if sign < 0 else 1)
provide(CopySign())

class Cos(LibFcn):
    name = prefix + "cos"
    sig = Sig([{"x": P.Double()}], P.Double())
    errcodeBase = 27090
    def __call__(self, state, scope, pos, paramTypes, x):
        if math.isinf(x) or math.isnan(x):
            return float("nan")
        else:
            return math.cos(x)
provide(Cos())

class CosH(LibFcn):
    name = prefix + "cosh"
    sig = Sig([{"x": P.Double()}], P.Double())
    errcodeBase = 27100
    def genpy(self, paramTypes, args, pos):
        return "math.cosh({0})".format(*args)
    def __call__(self, state, scope, pos, paramTypes, x):
        return math.cosh(x)
provide(CosH())

class Exp(LibFcn):
    name = prefix + "exp"
    sig = Sig([{"x": P.Double()}], P.Double())
    errcodeBase = 27110
    def genpy(self, paramTypes, args, pos):
        return "math.exp({0})".format(*args)
    def __call__(self, state, scope, pos, paramTypes, x):
        return math.exp(x)
provide(Exp())

class ExpM1(LibFcn):
    name = prefix + "expm1"
    sig = Sig([{"x": P.Double()}], P.Double())
    errcodeBase = 27120
    def genpy(self, paramTypes, args, pos):
        return "(math.exp({0}) - 1.0)".format(*args)
    def __call__(self, state, scope, pos, paramTypes, x):
        return math.exp(x) - 1.0
provide(ExpM1())

class Floor(LibFcn):
    name = prefix + "floor"
    sig = Sig([{"x": P.Double()}], P.Double())
    errcodeBase = 27130
#    def genpy(self, paramTypes, args, pos):
#        return "math.floor({0})".format(*args)
    def __call__(self, state, scope, pos, paramTypes, x):
        if math.isinf(x) or math.isnan(x):
            return x
        else:        
            return math.floor(x)
provide(Floor())

class Hypot(LibFcn):
    name = prefix + "hypot"
    sig = Sig([{"x": P.Double()}, {"y": P.Double()}], P.Double())
    errcodeBase = 27140
    def __call__(self, state, scope, pos, paramTypes, x, y):
        if math.isinf(x) or math.isinf(y):
            return float("inf")
        else:
            return math.sqrt(x**2 + y**2)
provide(Hypot())

class Ln(LibFcn):
    name = prefix + "ln"
    sig = Sig([{"x": P.Double()}], P.Double())
    errcodeBase = 27150
    def __call__(self, state, scope, pos, paramTypes, x):
        if x < 0.0:
            return float("nan")
        elif x == 0.0:
            return float("-inf")
        else:
            return math.log(x)
provide(Ln())

class Log10(LibFcn):
    name = prefix + "log10"
    sig = Sig([{"x": P.Double()}], P.Double())
    errcodeBase = 27160
    def __call__(self, state, scope, pos, paramTypes, x):
        if x < 0.0:
            return float("nan")
        elif x == 0.0:
            return float("-inf")
        else:
            return math.log10(x)
provide(Log10())

class Log(LibFcn):
    name = prefix + "log"
    sig = Sig([{"x": P.Double()}, {"base": P.Int()}], P.Double())
    errcodeBase = 27170
    def __call__(self, state, scope, pos, paramTypes, x, base):
        if base <= 0:
            raise PFARuntimeException("base must be positive", self.errcodeBase + 0, self.name, pos)
        elif math.isnan(x):
            return float("nan")
        elif x < 0.0:
            return float("nan")
        elif x == 0.0:
            return float("-inf")
        elif base == 1:
            if x < 1.0:
                return float("-inf")
            elif x == 1.0:
                return float("nan")
            elif x > 1.0:
                return float("inf")
        else:
            return math.log(x, base)
provide(Log())

class Ln1p(LibFcn):
    name = prefix + "ln1p"
    sig = Sig([{"x": P.Double()}], P.Double())
    errcodeBase = 27180
    def __call__(self, state, scope, pos, paramTypes, x):
        if x < -1.0:
            return float("nan")
        elif x == -1.0:
            return float("-inf")
        else:
            return math.log(1 + x)
provide(Ln1p())

class Round(LibFcn):
    name = prefix + "round"
    sig = Sigs([Sig([{"x": P.Float()}], P.Int()),
                Sig([{"x": P.Double()}], P.Long())])
    errcodeBase = 27190
    def __call__(self, state, scope, pos, paramTypes, x):
        out = None
        if math.isinf(x) or math.isnan(x):
            out = x
        else:
            out = math.floor(x + 0.5)
        return int(checkForOverflow(paramTypes[-1], out, self.errcodeBase + 0, self.errcodeBase + 1, self.name, pos))
provide(Round())

class RInt(LibFcn):
    name = prefix + "rint"
    sig = Sig([{"x": P.Double()}], P.Double())
    errcodeBase = 27200
    def __call__(self, state, scope, pos, paramTypes, x):
        if math.isinf(x) or math.isnan(x):
            return x
        up, down = math.ceil(x), math.floor(x)
        if up - x < x - down:
            return up
        elif up - x > x - down:
            return down
        elif int(up) % 2 == 0:
            return up
        else:
            return down
            
provide(RInt())

class Signum(LibFcn):
    name = prefix + "signum"
    sig = Sig([{"x": P.Double()}], P.Int())
    errcodeBase = 27210
    def __call__(self, state, scope, pos, paramTypes, x):
        if math.isnan(x):
            return 0
        elif x == 0:
            return 0
        elif x > 0:
            return 1
        else:
            return -1
provide(Signum())

class Sin(LibFcn):
    name = prefix + "sin"
    sig = Sig([{"x": P.Double()}], P.Double())
    errcodeBase = 27220
    def __call__(self, state, scope, pos, paramTypes, x):
        if math.isinf(x) or math.isnan(x):
            return float("nan")
        else:
            return math.sin(x)
provide(Sin())

class SinH(LibFcn):
    name = prefix + "sinh"
    sig = Sig([{"x": P.Double()}], P.Double())
    errcodeBase = 27230
    def genpy(self, paramTypes, args, pos):
        return "math.sinh({0})".format(*args)
    def __call__(self, state, scope, pos, paramTypes, x):
        return math.sinh(x)
provide(SinH())

class Sqrt(LibFcn):
    name = prefix + "sqrt"
    sig = Sig([{"x": P.Double()}], P.Double())
    errcodeBase = 27240
    def __call__(self, state, scope, pos, paramTypes, x):
        try:
            return math.sqrt(x)
        except ValueError:
            return float("nan")
provide(Sqrt())

class Tan(LibFcn):
    name = prefix + "tan"
    sig = Sig([{"x": P.Double()}], P.Double())
    errcodeBase = 27250
    def __call__(self, state, scope, pos, paramTypes, x):
        if math.isinf(x) or math.isnan(x):
            return float("nan")
        else:
            return math.tan(x)
provide(Tan())

class TanH(LibFcn):
    name = prefix + "tanh"
    sig = Sig([{"x": P.Double()}], P.Double())
    errcodeBase = 27260
    def genpy(self, paramTypes, args, pos):
        return "math.tanh({0})".format(*args)
    def __call__(self, state, scope, pos, paramTypes, x):
        return math.tanh(x)
provide(TanH())

#################################################################### special functions

# class SpecialErf(LibFcn):
#     name = prefix + "special.erf"
#     sig = Sig([{"x": P.Double()}], P.Double())
#     a1 =  0.254829592
#     a2 = -0.284496736
#     a3 =  1.421413741
#     a4 = -1.453152027
#     a5 =  1.061405429
#     p  =  0.3275911
#     def __call__(self, state, scope, pos, paramTypes, x):
#         try:
#             return math.erf(x)
#         except AttributeError:
#             sign = -1.0 if x < 0 else 1.0
#             x = abs(x)
#             t = 1.0 / (1.0 + x * self.p)
#             y = 1.0 - (((((self.a5*t + self.a4)*t + self.a3)*t + self.a2)*t + self.a1)*t * math.exp(-x**2))
#             return sign * y    # erf(-x) = -erf(x)

# provide(SpecialErf())




#!/usr/bin/env python

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from os import path, walk
from setuptools import setup

import titus.version

### To run the tests:
# python setup.py test

### To install on a machine:
# sudo python setup.py install

### To install in a home directory (~/lib):
# python setup.py install --home=~
LONG_DESCRIPTION = open(path.join(path.dirname(__file__), 'README.md')).read()

setup(name="titus2",
      version=titus.version.__version__,
      author="Ankit Mahato",
      author_email="ankit@realworldpython.guide",
      keywords='pfa scoring inference pmml titus hadrian machine learning data mining',
      packages=["titus",
                "titus.producer",
                "titus.lib",
                "titus.lib.model",
                "titus.lib.prob",
                "titus.lib.stat",
                "titus.inspector"],
      scripts = ["scripts/pfainspector", "scripts/pfachain", "scripts/pfaexternalize", "scripts/pfarandom", "scripts/pfasize"],
      description="Python 3 implementation of Portable Format for Analytics (PFA): producer, converter, and consumer.",
      long_description=LONG_DESCRIPTION,
      long_description_content_type="text/markdown",
      url="https://realworldpython.guide/titus2/",
      project_urls={
            "Source Code": "https://github.com/animator/titus2",
            "Bug Tracker": "https://github.com/animator/titus2/issues",
            "Documentation": "https://realworldpython.guide/titus2/",
      },
      test_suite="test",
      install_requires=["avro-python3>=1.8.2", "ply>=3.11", "pyyaml>=5.1.2", "numpy>=1.15.0", "pytz>=2019.1"],
      tests_require=["avro-python3>=1.8.2", "ply>=3.11", "pyyaml>=5.1.2", "numpy>=1.15.0", "pytz>=2019.1"],
      python_requires='>=3.4',
      classifiers=[  
            'Development Status :: 5 - Production/Stable', 
            'Intended Audience :: Developers',
            'Intended Audience :: Education',
            'Intended Audience :: Information Technology',
            'Intended Audience :: Science/Research',
            'Topic :: Scientific/Engineering',
            'Topic :: Scientific/Engineering :: Artificial Intelligence',
            'Topic :: Software Development',
            'Topic :: Software Development :: Libraries',
            'Topic :: Software Development :: Libraries :: Python Modules',
            'License :: OSI Approved :: Apache Software License',
            'Programming Language :: Python :: 3 :: Only',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Natural Language :: English',
      ],      
      )

### details of dependencies:
# 
# Tested in Python 3.4-3.8.
# Will not work in Python 2.x.
# 
# avro-python3 is required; it is an integral part of PFA.
# ply is required; it is used to parse PrettyPFA and Inspector commandlines.
# 
# PyYAML is an optional dependency; it is only used by the titus.reader.yamlToAst function.
# Numpy is an optional dependency; it is only used by the "interp", "la", "stat.test", and "model.reg" PFA libraries, as well as Titus producers.
# pytz is an optional dependency; it is only used by the "time" PFA library.



#!/usr/bin/env python

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import io
import json

import avro.schema
from avro.io import BinaryEncoder, DatumWriter

from titus.util import untagUnion, bytesToString
from titus.datatype import schemaToAvroType
from titus.datatype import jsonNodeToAvroType
from titus.fcn import Fcn
from titus.fcn import LibFcn
from titus.signature import Sig
from titus.signature import Sigs
from titus.datatype import *
from titus.errors import *
from titus.lib.core import INT_MIN_VALUE
from titus.lib.core import INT_MAX_VALUE
from titus.lib.core import LONG_MIN_VALUE
from titus.lib.core import LONG_MAX_VALUE
from titus.lib.core import FLOAT_MIN_VALUE
from titus.lib.core import FLOAT_MAX_VALUE
from titus.lib.core import DOUBLE_MIN_VALUE
from titus.lib.core import DOUBLE_MAX_VALUE
from titus.datatype import jsonEncoder
import titus.P as P

provides = {}
def provide(fcn):
    provides[fcn.name] = fcn

prefix = "cast."

#################################################################### wrap-around arithmetic

def bitsToMax(x): return 2**x

def doUnsigned(x, bits):
    maximum = bitsToMax(bits)
    if x < 0:
        y = x + maximum * int(math.ceil(-float(x) / maximum))
    else:
        y = x
    return y % maximum

class ToSigned(LibFcn):
    name = prefix + "signed"
    sig = Sig([{"x": P.Long()}, {"bits": P.Int()}], P.Long())
    errcodeBase = 17000
    def __call__(self, state, scope, pos, paramTypes, x, bits):
        if bits < 2 or bits > 64:
            raise PFARuntimeException("unrepresentable unsigned number", self.errcodeBase + 0, self.name, pos)
        y = doUnsigned(x, bits)
        maximum = bitsToMax(bits - 1)
        if y > maximum - 1:
            return y - 2*maximum
        else:
            return y
provide(ToSigned())

class ToUnsigned(LibFcn):
    name = prefix + "unsigned"
    sig = Sig([{"x": P.Long()}, {"bits": P.Int()}], P.Long())
    errcodeBase = 17010
    def __call__(self, state, scope, pos, paramTypes, x, bits):
        if bits < 1 or bits > 63:
            raise PFARuntimeException("unrepresentable unsigned number", self.errcodeBase + 0, self.name, pos)
        return doUnsigned(x, bits)
provide(ToUnsigned())

#################################################################### number precisions

class ToInt(LibFcn):
    name = prefix + "int"
    sig = Sigs([Sig([{"x": P.Int()}], P.Int()),
                Sig([{"x": P.Long()}], P.Int()),
                Sig([{"x": P.Float()}], P.Int()),
                Sig([{"x": P.Double()}], P.Int())])
    errcodeBase = 17020
    def __call__(self, state, scope, pos, paramTypes, x):
        try:
            if isinstance(x, float):
                if math.isnan(x):
                    raise OverflowError
                else:
                    out = int(math.floor(x + 0.5))
            else:
                out = x
            if INT_MIN_VALUE <= out <= INT_MAX_VALUE:
                return out
            else:
                raise OverflowError
        except OverflowError:
            raise PFARuntimeException("int overflow", self.errcodeBase + 0, self.name, pos)
provide(ToInt())

class ToLong(LibFcn):
    name = prefix + "long"
    sig = Sigs([Sig([{"x": P.Int()}], P.Long()),
                Sig([{"x": P.Long()}], P.Long()),
                Sig([{"x": P.Float()}], P.Long()),
                Sig([{"x": P.Double()}], P.Long())])
    errcodeBase = 17030
    def __call__(self, state, scope, pos, paramTypes, x):
        try:
            if isinstance(x, float):
                if math.isnan(x):
                    raise OverflowError
                else:
                    out = int(math.floor(x + 0.5))
            else:
                out = x
            if LONG_MIN_VALUE <= out <= LONG_MAX_VALUE:
                return out
            else:
                raise OverflowError
        except OverflowError:
            raise PFARuntimeException("long overflow", self.errcodeBase + 0, self.name, pos)
provide(ToLong())

class ToFloat(LibFcn):
    name = prefix + "float"
    sig = Sigs([Sig([{"x": P.Int()}], P.Float()),
                Sig([{"x": P.Long()}], P.Float()),
                Sig([{"x": P.Float()}], P.Float()),
                Sig([{"x": P.Double()}], P.Float())])
    errcodeBase = 17040
    def genpy(self, paramTypes, args, pos):
        return "float({0})".format(*args)
    def __call__(self, state, scope, pos, paramTypes, x):
        return float(x)
provide(ToFloat())

class ToDouble(LibFcn):
    name = prefix + "double"
    sig = Sigs([Sig([{"x": P.Int()}], P.Double()),
                Sig([{"x": P.Long()}], P.Double()),
                Sig([{"x": P.Float()}], P.Double()),
                Sig([{"x": P.Double()}], P.Double())])
    errcodeBase = 17050
    def genpy(self, paramTypes, args, pos):
        return "float({0})".format(*args)
    def __call__(self, state, scope, pos, paramTypes, x):
        return float(x)
provide(ToDouble())

#################################################################### fanouts

def fanoutEnum(x, symbols):
    return [x == s for s in symbols]

def fanoutString(x, dictionary, outOfRange):
    out = [x == s for s in dictionary]
    if outOfRange:
        return out + [x not in dictionary]
    else:
        return out

def fanoutInt(x, minimum, maximum, outOfRange):
    out = [x == i for i in range(minimum, maximum)]
    if outOfRange:
        return out + [x < minimum or x >= maximum]
    else:
        return out

class FanoutBoolean(LibFcn):
    name = prefix + "fanoutBoolean"
    sig = Sigs([Sig([{"x": P.WildEnum("A")}], P.Array(P.Boolean())),
                Sig([{"x": P.String()}, {"dictionary": P.Array(P.String())}, {"outOfRange": P.Boolean()}], P.Array(P.Boolean())),
                Sig([{"x": P.Int()}, {"minimum": P.Int()}, {"maximum": P.Int()}, {"outOfRange": P.Boolean()}], P.Array(P.Boolean()))])
    errcodeBase = 17060
    def __call__(self, state, scope, pos, paramTypes, x, *args):
        if len(args) == 0:
            return fanoutEnum(x, paramTypes[0]["symbols"])
        elif len(args) == 2:
            if len(args[0]) != len(set(args[0])):
                raise PFARuntimeException("non-distinct values in dictionary", self.errcodeBase + 0, self.name, pos)
            return fanoutString(x, args[0], args[1])
        elif len(args) == 3:
            return fanoutInt(x, args[0], args[1], args[2])
provide(FanoutBoolean())

class FanoutInt(LibFcn):
    name = prefix + "fanoutInt"
    sig = Sigs([Sig([{"x": P.WildEnum("A")}], P.Array(P.Int())),
                Sig([{"x": P.String()}, {"dictionary": P.Array(P.String())}, {"outOfRange": P.Boolean()}], P.Array(P.Int())),
                Sig([{"x": P.Int()}, {"minimum": P.Int()}, {"maximum": P.Int()}, {"outOfRange": P.Boolean()}], P.Array(P.Int()))])
    errcodeBase = 17070
    def __call__(self, state, scope, pos, paramTypes, x, *args):
        if len(args) == 0:
            return [1 if y else 0 for y in fanoutEnum(x, paramTypes[0]["symbols"])]
        elif len(args) == 2:
            if len(args[0]) != len(set(args[0])):
                raise PFARuntimeException("non-distinct values in dictionary", self.errcodeBase + 0, self.name, pos)
            return [1 if y else 0 for y in fanoutString(x, args[0], args[1])]
        elif len(args) == 3:
            return [1 if y else 0 for y in fanoutInt(x, args[0], args[1], args[2])]
provide(FanoutInt())

class FanoutLong(LibFcn):
    name = prefix + "fanoutLong"
    sig = Sigs([Sig([{"x": P.WildEnum("A")}], P.Array(P.Long())),
                Sig([{"x": P.String()}, {"dictionary": P.Array(P.String())}, {"outOfRange": P.Boolean()}], P.Array(P.Long())),
                Sig([{"x": P.Int()}, {"minimum": P.Int()}, {"maximum": P.Int()}, {"outOfRange": P.Boolean()}], P.Array(P.Long()))])
    errcodeBase = 17080
    def __call__(self, state, scope, pos, paramTypes, x, *args):
        if len(args) == 0:
            return [1 if y else 0 for y in fanoutEnum(x, paramTypes[0]["symbols"])]
        elif len(args) == 2:
            if len(args[0]) != len(set(args[0])):
                raise PFARuntimeException("non-distinct values in dictionary", self.errcodeBase + 0, self.name, pos)
            return [1 if y else 0 for y in fanoutString(x, args[0], args[1])]
        elif len(args) == 3:
            return [1 if y else 0 for y in fanoutInt(x, args[0], args[1], args[2])]
provide(FanoutLong())

class FanoutFloat(LibFcn):
    name = prefix + "fanoutFloat"
    sig = Sigs([Sig([{"x": P.WildEnum("A")}], P.Array(P.Float())),
                Sig([{"x": P.String()}, {"dictionary": P.Array(P.String())}, {"outOfRange": P.Boolean()}], P.Array(P.Float())),
                Sig([{"x": P.Int()}, {"minimum": P.Int()}, {"maximum": P.Int()}, {"outOfRange": P.Boolean()}], P.Array(P.Float()))])
    errcodeBase = 17090
    def __call__(self, state, scope, pos, paramTypes, x, *args):
        if len(args) == 0:
            return [1.0 if y else 0.0 for y in fanoutEnum(x, paramTypes[0]["symbols"])]
        elif len(args) == 2:
            if len(args[0]) != len(set(args[0])):
                raise PFARuntimeException("non-distinct values in dictionary", self.errcodeBase + 0, self.name, pos)
            return [1.0 if y else 0.0 for y in fanoutString(x, args[0], args[1])]
        elif len(args) == 3:
            return [1.0 if y else 0.0 for y in fanoutInt(x, args[0], args[1], args[2])]
provide(FanoutFloat())

class FanoutDouble(LibFcn):
    name = prefix + "fanoutDouble"
    sig = Sigs([Sig([{"x": P.WildEnum("A")}], P.Array(P.Double())),
                Sig([{"x": P.String()}, {"dictionary": P.Array(P.String())}, {"outOfRange": P.Boolean()}], P.Array(P.Double())),
                Sig([{"x": P.Int()}, {"minimum": P.Int()}, {"maximum": P.Int()}, {"outOfRange": P.Boolean()}], P.Array(P.Double()))])
    errcodeBase = 17100
    def __call__(self, state, scope, pos, paramTypes, x, *args):
        if len(args) == 0:
            return [1.0 if y else 0.0 for y in fanoutEnum(x, paramTypes[0]["symbols"])]
        elif len(args) == 2:
            if len(args[0]) != len(set(args[0])):
                raise PFARuntimeException("non-distinct values in dictionary", self.errcodeBase + 0, self.name, pos)
            return [1.0 if y else 0.0 for y in fanoutString(x, args[0], args[1])]
        elif len(args) == 3:
            return [1.0 if y else 0.0 for y in fanoutInt(x, args[0], args[1], args[2])]
provide(FanoutDouble())

#################################################################### serialize

class CastAvro(LibFcn):
    name = prefix + "avro"
    sig = Sig([{"x": P.Wildcard("A")}], P.Bytes())
    errcodeBase = 17110
    def __call__(self, state, scope, pos, paramTypes, x):
        schema = avro.schema.Parse(json.dumps(paramTypes[0]))
        x = untagUnion(x, paramTypes[0])
        bytes_io = io.BytesIO()
        writer = DatumWriter(schema)
        writer.write(x, BinaryEncoder(bytes_io))
        bytes_io.flush()
        return bytesToString(bytes_io.getvalue())
provide(CastAvro())

class CastJson(LibFcn):
    name = prefix + "json"
    sig = Sig([{"x": P.Wildcard("A")}], P.String())
    errcodeBase = 17120
    def __call__(self, state, scope, pos, paramTypes, x):
        return json.dumps(jsonEncoder(jsonNodeToAvroType(paramTypes[0]), x), separators=(",", ":"))
provide(CastJson())



#!/usr/bin/env python

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from titus.fcn import Fcn
from titus.fcn import LibFcn
from titus.signature import Sig
from titus.signature import Sigs
from titus.datatype import *
from titus.errors import AvroException
import titus.P as P

provides = {}
def provide(fcn):
    provides[fcn.name] = fcn

prefix = "enum."

class ToString(LibFcn):
    name = prefix + "toString"
    sig = Sig([{"x": P.WildEnum("A")}], P.String())
    errcodeBase = 19000
    def __call__(self, state, scope, pos, paramTypes, x):
        if x in paramTypes[0]["symbols"]:
            return x
        else:
            raise AvroException("\"{0}\" is not in enum {1} ({2})".format(x, paramTypes[0]["name"], paramTypes[0]["symbols"]))
provide(ToString())

class ToInt(LibFcn):
    name = prefix + "toInt"
    sig = Sig([{"x": P.WildEnum("A")}], P.Int())
    errcodeBase = 19010
    def __call__(self, state, scope, pos, paramTypes, x):
        try:
            return paramTypes[0]["symbols"].index(x)
        except ValueError:
            raise AvroException("\"{0}\" is not in enum {1} ({2})".format(x, paramTypes[0]["name"], paramTypes[0]["symbols"]))
provide(ToInt())

class NumSymbols(LibFcn):
    name = prefix + "numSymbols"
    sig = Sig([{"x": P.WildEnum("A")}], P.Int())
    errcodeBase = 19020
    def __call__(self, state, scope, pos, paramTypes, x):
        return len(paramTypes[0]["symbols"])
provide(NumSymbols())



#!/usr/bin/env python

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import itertools
import json

from titus.fcn import Fcn
from titus.fcn import LibFcn
from titus.signature import Sig
from titus.signature import Sigs
from titus.datatype import *
from titus.errors import *
from titus.util import callfcn, negativeIndex, startEnd
from titus.lib.core import INT_MIN_VALUE
from titus.lib.core import INT_MAX_VALUE
from titus.lib.core import LONG_MIN_VALUE
from titus.lib.core import LONG_MAX_VALUE
import titus.P as P
from functools import reduce, cmp_to_key

provides = {}
def provide(fcn):
    provides[fcn.name] = fcn

prefix = "a."

anyNumber = set([AvroInt(), AvroLong(), AvroFloat(), AvroDouble()])

def toCmp(state, scope, lessThan):
    return lambda a, b: -1 if callfcn(state, scope, lessThan, [a, b]) else 1 if callfcn(state, scope, lessThan, [b, a]) else 0

def toLt(state, scope, lessThan):
    return lambda a, b: callfcn(state, scope, lessThan, [a, b])

def checkRange(length, index, code, fcnName, pos):
    if index < 0 or index >= length:
        raise PFARuntimeException("index out of range", code, fcnName, pos)

#################################################################### basic access

class Len(LibFcn):
    name = prefix + "len"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}], P.Int())
    errcodeBase = 15000
    def genpy(self, paramTypes, args, pos):
        return "len({0})".format(*args)
    def __call__(self, state, scope, pos, paramTypes, a):
        return len(a)
provide(Len())

class Subseq(LibFcn):
    name = prefix + "subseq"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"start": P.Int()}, {"end": P.Int()}], P.Array(P.Wildcard("A")))
    errcodeBase = 15010
    def genpy(self, paramTypes, args, pos):
        return "{0}[{1}:{2}]".format(*args)
    def __call__(self, state, scope, pos, paramTypes, a, start, end):
        return a[start:end]
provide(Subseq())

class Head(LibFcn):
    name = prefix + "head"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}], P.Wildcard("A"))
    errcodeBase = 15020
    def __call__(self, state, scope, pos, paramTypes, a):
        if len(a) == 0:
            raise PFARuntimeException("empty array", self.errcodeBase + 0, self.name, pos)
        else:
            return a[0]
provide(Head())

class Tail(LibFcn):
    name = prefix + "tail"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}], P.Array(P.Wildcard("A")))
    errcodeBase = 15030
    def __call__(self, state, scope, pos, paramTypes, a):
        if len(a) == 0:
            raise PFARuntimeException("empty array", self.errcodeBase + 0, self.name, pos)
        else:
            return a[1:]
provide(Tail())

class Last(LibFcn):
    name = prefix + "last"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}], P.Wildcard("A"))
    errcodeBase = 15040
    def __call__(self, state, scope, pos, paramTypes, a):
        if len(a) == 0:
            raise PFARuntimeException("empty array", self.errcodeBase + 0, self.name, pos)
        else:
            return a[-1]
provide(Last())

class Init(LibFcn):
    name = prefix + "init"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}], P.Array(P.Wildcard("A")))
    errcodeBase = 15050
    def __call__(self, state, scope, pos, paramTypes, a):
        if len(a) == 0:
            raise PFARuntimeException("empty array", self.errcodeBase + 0, self.name, pos)
        else:
            return a[:-1]
provide(Init())

class SubseqTo(LibFcn):
    name = prefix + "subseqto"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"start": P.Int()}, {"end": P.Int()}, {"replacement": P.Array(P.Wildcard("A"))}], P.Array(P.Wildcard("A")))
    errcodeBase = 15060
    def __call__(self, state, scope, pos, paramTypes, a, start, end, replacement):
        normStart, normEnd = startEnd(len(a), start, end)
        before = a[:normStart]
        after = a[normEnd:]
        return before + replacement + after
provide(SubseqTo())

#################################################################### searching

class Contains(LibFcn):
    name = prefix + "contains"
    sig = Sigs([Sig([{"haystack": P.Array(P.Wildcard("A"))}, {"needle": P.Array(P.Wildcard("A"))}], P.Boolean()),
                Sig([{"haystack": P.Array(P.Wildcard("A"))}, {"needle": P.Wildcard("A")}], P.Boolean()),
                Sig([{"haystack": P.Array(P.Wildcard("A"))}, {"needle": P.Fcn([P.Wildcard("A")], P.Boolean())}], P.Boolean())])
    errcodeBase = 15070
    def __call__(self, state, scope, pos, paramTypes, haystack, needle):
        if isinstance(needle, (list, tuple)):
            for start in range(len(haystack) - len(needle) + 1):
                if needle == haystack[start:(start + len(needle))]:
                    return True
            return False
        elif callable(needle):
            for item in haystack:
                if callfcn(state, scope, needle, [item]):
                    return True
            return False
        else:
            try:
                haystack.index(needle)
            except ValueError:
                return False
            else:
                return True
provide(Contains())

class Count(LibFcn):
    name = prefix + "count"
    sig = Sigs([Sig([{"haystack": P.Array(P.Wildcard("A"))}, {"needle": P.Array(P.Wildcard("A"))}], P.Int()),
                Sig([{"haystack": P.Array(P.Wildcard("A"))}, {"needle": P.Wildcard("A")}], P.Int()),
                Sig([{"a": P.Array(P.Wildcard("A"))}, {"predicate": P.Fcn([P.Wildcard("A")], P.Boolean())}], P.Int())])
    errcodeBase = 15080
    def __call__(self, state, scope, pos, paramTypes, haystack, needle):
        if len(haystack) == 0:
            return 0
        else:
            if isinstance(needle, (list, tuple)):
                if len(needle) == 0:
                    return 0
                else:
                    count = 0
                    for start in range(len(haystack) - len(needle) + 1):
                        if needle == haystack[start:(start + len(needle))]:
                            count += 1
                    return count
            elif callable(needle):
                count = 0
                for item in haystack:
                    if callfcn(state, scope, needle, [item]):
                        count += 1
                return count
            else:
                return haystack.count(needle)
provide(Count())

class Index(LibFcn):
    name = prefix + "index"
    sig = Sigs([Sig([{"haystack": P.Array(P.Wildcard("A"))}, {"needle": P.Array(P.Wildcard("A"))}], P.Int()),
                Sig([{"haystack": P.Array(P.Wildcard("A"))}, {"needle": P.Wildcard("A")}], P.Int()),
                Sig([{"haystack": P.Array(P.Wildcard("A"))}, {"needle": P.Fcn([P.Wildcard("A")], P.Boolean())}], P.Int())])
    errcodeBase = 15090
    def __call__(self, state, scope, pos, paramTypes, haystack, needle):
        if isinstance(needle, (list, tuple)):
            for start in range(len(haystack) - len(needle) + 1):
                if needle == haystack[start:(start + len(needle))]:
                    return start
            return -1
        elif callable(needle):
            for index, item in enumerate(haystack):
                if callfcn(state, scope, needle, [item]):
                    return index
            return -1
        else:
            try:
                return haystack.index(needle)
            except ValueError:
                return -1
provide(Index())

class RIndex(LibFcn):
    name = prefix + "rindex"
    sig = Sigs([Sig([{"haystack": P.Array(P.Wildcard("A"))}, {"needle": P.Array(P.Wildcard("A"))}], P.Int()),
                Sig([{"haystack": P.Array(P.Wildcard("A"))}, {"needle": P.Wildcard("A")}], P.Int()),
                Sig([{"haystack": P.Array(P.Wildcard("A"))}, {"needle": P.Fcn([P.Wildcard("A")], P.Boolean())}], P.Int())])
    errcodeBase = 15100
    def __call__(self, state, scope, pos, paramTypes, haystack, needle):
        if isinstance(needle, (list, tuple)):
            for start in range(len(haystack) - len(needle), -1, -1):
                if needle == haystack[start:(start + len(needle))]:
                    return start
            return -1
        elif callable(needle):
            for index, item in enumerate(reversed(haystack)):
                if callfcn(state, scope, needle, [item]):
                    return len(haystack) - 1 - index
            return -1
        else:
            for index in range(len(haystack) - 1, -1, -1):
                if needle == haystack[index]:
                    return index
            return -1
provide(RIndex())

class StartsWith(LibFcn):
    name = prefix + "startswith"
    sig = Sigs([Sig([{"haystack": P.Array(P.Wildcard("A"))}, {"needle": P.Array(P.Wildcard("A"))}], P.Boolean()),
                Sig([{"haystack": P.Array(P.Wildcard("A"))}, {"needle": P.Wildcard("A")}], P.Boolean())])
    errcodeBase = 15110
    def __call__(self, state, scope, pos, paramTypes, haystack, needle):
        if isinstance(needle, (list, tuple)):
            return needle == haystack[:len(needle)]
        else:
            if len(haystack) == 0:
                return False
            else:
                return needle == haystack[0]
provide(StartsWith())

class EndsWith(LibFcn):
    name = prefix + "endswith"
    sig = Sigs([Sig([{"haystack": P.Array(P.Wildcard("A"))}, {"needle": P.Array(P.Wildcard("A"))}], P.Boolean()),
                Sig([{"haystack": P.Array(P.Wildcard("A"))}, {"needle": P.Wildcard("A")}], P.Boolean())])
    errcodeBase = 15120
    def __call__(self, state, scope, pos, paramTypes, haystack, needle):
        if isinstance(needle, (list, tuple)):
            if len(needle) == 0:
                return True
            else:
                return needle == haystack[-len(needle):]
        else:
            if len(haystack) == 0:
                return False
            else:
                return needle == haystack[-1]
provide(EndsWith())

#################################################################### manipulation

class Concat(LibFcn):
    name = prefix + "concat"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"b": P.Array(P.Wildcard("A"))}], P.Array(P.Wildcard("A")))
    errcodeBase = 15130
    def genpy(self, paramTypes, args, pos):
        return "({0} + {1})".format(*args)
    def __call__(self, state, scope, pos, paramTypes, a, b):
        return a + b
provide(Concat())

class Append(LibFcn):
    name = prefix + "append"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"item": P.Wildcard("A")}], P.Array(P.Wildcard("A")))
    errcodeBase = 15140
    def genpy(self, paramTypes, args, pos):
        return "({0} + [{1}])".format(*args)
    def __call__(self, state, scope, pos, paramTypes, a, item):
        return a + [item]
provide(Append())

class Cycle(LibFcn):
    name = prefix + "cycle"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"item": P.Wildcard("A")}, {"maxLength": P.Int()}], P.Array(P.Wildcard("A")))
    errcodeBase = 15150
    def __call__(self, state, scope, pos, paramTypes, a, item, maxLength):
        if maxLength < 0:
            raise PFARuntimeException("maxLength out of range", self.errcodeBase + 0, self.name, pos)
        elif maxLength == 0:
            return []
        else:
            out = a + [item]
            return out[-maxLength:]
provide(Cycle())

class Insert(LibFcn):
    name = prefix + "insert"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"index": P.Int()}, {"item": P.Wildcard("A")}], P.Array(P.Wildcard("A")))
    errcodeBase = 15160
    def __call__(self, state, scope, pos, paramTypes, a, index, item):
        normIndex = negativeIndex(len(a), index)
        checkRange(len(a), normIndex, self.errcodeBase + 0, self.name, pos)
        before = a[:normIndex]
        after = a[normIndex:]
        return before + [item] + after
provide(Insert())

class Replace(LibFcn):
    name = prefix + "replace"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"index": P.Int()}, {"item": P.Wildcard("A")}], P.Array(P.Wildcard("A")))
    errcodeBase = 15170
    def __call__(self, state, scope, pos, paramTypes, a, index, item):
        normIndex = negativeIndex(len(a), index)
        checkRange(len(a), normIndex, self.errcodeBase + 0, self.name, pos)
        before = a[:normIndex]
        after = a[(normIndex + 1):]
        return before + [item] + after
provide(Replace())

class Remove(LibFcn):
    name = prefix + "remove"
    sig = Sigs([Sig([{"a": P.Array(P.Wildcard("A"))}, {"start": P.Int()}, {"end": P.Int()}], P.Array(P.Wildcard("A"))),
                Sig([{"a": P.Array(P.Wildcard("A"))}, {"index": P.Int()}], P.Array(P.Wildcard("A")))])
    errcodeBase = 15180
    def __call__(self, state, scope, pos, paramTypes, a, *where):
        if len(where) == 2:
            start, end = where
            normStart, normEnd = startEnd(len(a), start, end)
            before = a[:normStart]
            after = a[normEnd:]
            return before + after
        elif len(where) == 1:
            index, = where
            normIndex = negativeIndex(len(a), index)
            checkRange(len(a), normIndex, self.errcodeBase + 0, self.name, pos)
            before = a[:normIndex]
            after = a[(normIndex + 1):]
            return before + after
provide(Remove())

class Rotate(LibFcn):
    name = prefix + "rotate"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"steps": P.Int()}], P.Array(P.Wildcard("A")))
    errcodeBase = 15190
    def __call__(self, state, scope, pos, paramTypes, a, steps):
        if steps < 0:
            raise PFARuntimeException("steps out of range", self.errcodeBase + 0, self.name, pos)
        elif steps == 0 or len(a) == 0:
            return a
        else:
            index = steps % len(a)
            left, right = a[:index], a[index:]
            return right + left
provide(Rotate())

#################################################################### reordering

class Sort(LibFcn):
    name = prefix + "sort"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}], P.Array(P.Wildcard("A")))
    errcodeBase = 15200
    def __call__(self, state, scope, pos, paramTypes, a):
        return sorted(a, key=cmp_to_key(lambda x, y: compare(jsonNodeToAvroType(paramTypes[0]).items, x, y)))
provide(Sort())

class SortLT(LibFcn):
    name = prefix + "sortLT"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"lessThan": P.Fcn([P.Wildcard("A"), P.Wildcard("A")], P.Boolean())}], P.Array(P.Wildcard("A")))
    errcodeBase = 15210
    def __call__(self, state, scope, pos, paramTypes, a, lessThan):
        return sorted(a, key=cmp_to_key(toCmp(state, scope, lessThan)))
provide(SortLT())

class Shuffle(LibFcn):
    name = prefix + "shuffle"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}], P.Array(P.Wildcard("A")))
    errcodeBase = 15220
    def __call__(self, state, scope, pos, paramTypes, a):
        out = list(a)
        state.rand.shuffle(out)
        return out
provide(Shuffle())

class Reverse(LibFcn):
    name = prefix + "reverse"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}], P.Array(P.Wildcard("A")))
    errcodeBase = 15230
    def __call__(self, state, scope, pos, paramTypes, a):
        return list(reversed(a))
provide(Reverse())

#################################################################### extreme values

def highestN(a, n, lt):
    out = []
    for x in a:
        found = False
        for index, best in enumerate(out):
            if lt(best, x):
                out.insert(index, x)
                found = True
                break
        if not found:
            out.append(x)
        if len(out) > n:
            out.pop()
    return out

def lowestN(a, n, lt):
    out = []
    for x in a:
        found = False
        for index, best in enumerate(out):
            if lt(x, best):
                out.insert(index, x)
                found = True
                break
        if not found:
            out.append(x)
        if len(out) > n:
            out.pop()
    return out

def argHighestN(a, n, lt):
    out = []
    for i, x in enumerate(a):
        found = False
        for index, (besti, bestx) in enumerate(out):
            if lt(bestx, x):
                ind = index
                while ind <= len(out) and not lt(out[ind][1], x) and not lt(x, out[ind][1]) and out[ind][0] > i:
                    ind += 1
                out.insert(ind, (i, x))
                found = True
                break
        if not found:
            out.append((i, x))
        if len(out) > n:
            out.pop()
    return [i for i, x in out]

def argLowestN(a, n, lt):
    out = []
    for i, x in enumerate(a):
        found = False
        for index, (besti, bestx) in enumerate(out):
            if lt(x, bestx):
                ind = index
                while ind <= len(out) and not lt(out[ind][1], x) and not lt(x, out[ind][1]) and out[ind][0] > i:
                    ind += 1
                out.insert(ind, (i, x))
                found = True
                break
        if not found:
            out.append((i, x))
        if len(out) > n:
            out.pop()
    return [i for i, x in out]

class Max(LibFcn):
    name = prefix + "max"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}], P.Wildcard("A"))
    errcodeBase = 15240
    def __call__(self, state, scope, pos, paramTypes, a):
        if len(a) == 0:
            raise PFARuntimeException("empty array", self.errcodeBase + 0, self.name, pos)
        else:
            return highestN(a, 1, lambda x, y: compare(jsonNodeToAvroType(paramTypes[0]).items, x, y) < 0)[0]
provide(Max())

class Min(LibFcn):
    name = prefix + "min"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}], P.Wildcard("A"))
    errcodeBase = 15250
    def __call__(self, state, scope, pos, paramTypes, a):
        if len(a) == 0:
            raise PFARuntimeException("empty array", self.errcodeBase + 0, self.name, pos)
        else:
            return lowestN(a, 1, lambda x, y: compare(jsonNodeToAvroType(paramTypes[0]).items, x, y) < 0)[0]
provide(Min())

class MaxLT(LibFcn):
    name = prefix + "maxLT"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"lessThan": P.Fcn([P.Wildcard("A"), P.Wildcard("A")], P.Boolean())}], P.Wildcard("A"))
    errcodeBase = 15260
    def __call__(self, state, scope, pos, paramTypes, a, lessThan):
        if len(a) == 0:
            raise PFARuntimeException("empty array", self.errcodeBase + 0, self.name, pos)
        else:
            return highestN(a, 1, toLt(state, scope, lessThan))[0]
provide(MaxLT())

class MinLT(LibFcn):
    name = prefix + "minLT"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"lessThan": P.Fcn([P.Wildcard("A"), P.Wildcard("A")], P.Boolean())}], P.Wildcard("A"))
    errcodeBase = 15270
    def __call__(self, state, scope, pos, paramTypes, a, lessThan):
        if len(a) == 0:
            raise PFARuntimeException("empty array", self.errcodeBase + 0, self.name, pos)
        else:
            return lowestN(a, 1, toLt(state, scope, lessThan))[0]
provide(MinLT())

class MaxN(LibFcn):
    name = prefix + "maxN"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"n": P.Int()}], P.Array(P.Wildcard("A")))
    errcodeBase = 15280
    def __call__(self, state, scope, pos, paramTypes, a, n):
        if len(a) == 0:
            raise PFARuntimeException("empty array", self.errcodeBase + 0, self.name, pos)
        elif n < 0:
            raise PFARuntimeException("n < 0", self.errcodeBase + 1, self.name, pos)
        else:
            return highestN(a, n, lambda x, y: compare(jsonNodeToAvroType(paramTypes[0]).items, x, y) < 0)
provide(MaxN())

class MinN(LibFcn):
    name = prefix + "minN"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"n": P.Int()}], P.Array(P.Wildcard("A")))
    errcodeBase = 15290
    def __call__(self, state, scope, pos, paramTypes, a, n):
        if len(a) == 0:
            raise PFARuntimeException("empty array", self.errcodeBase + 0, self.name, pos)
        elif n < 0:
            raise PFARuntimeException("n < 0", self.errcodeBase + 1, self.name, pos)
        else:
            return lowestN(a, n, lambda x, y: compare(jsonNodeToAvroType(paramTypes[0]).items, x, y) < 0)
provide(MinN())

class MaxNLT(LibFcn):
    name = prefix + "maxNLT"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"n": P.Int()}, {"lessThan": P.Fcn([P.Wildcard("A"), P.Wildcard("A")], P.Boolean())}], P.Array(P.Wildcard("A")))
    errcodeBase = 15300
    def __call__(self, state, scope, pos, paramTypes, a, n, lessThan):
        if len(a) == 0:
            raise PFARuntimeException("empty array", self.errcodeBase + 0, self.name, pos)
        elif n < 0:
            raise PFARuntimeException("n < 0", self.errcodeBase + 1, self.name, pos)
        else:
            return highestN(a, n, toLt(state, scope, lessThan))
provide(MaxNLT())

class MinNLT(LibFcn):
    name = prefix + "minNLT"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"n": P.Int()}, {"lessThan": P.Fcn([P.Wildcard("A"), P.Wildcard("A")], P.Boolean())}], P.Array(P.Wildcard("A")))
    errcodeBase = 15310
    def __call__(self, state, scope, pos, paramTypes, a, n, lessThan):
        if len(a) == 0:
            raise PFARuntimeException("empty array", self.errcodeBase + 0, self.name, pos)
        elif n < 0:
            raise PFARuntimeException("n < 0", self.errcodeBase + 1, self.name, pos)
        else:
            return lowestN(a, n, toLt(state, scope, lessThan))
provide(MinNLT())

class Argmax(LibFcn):
    name = prefix + "argmax"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}], P.Int())
    errcodeBase = 15320
    def __call__(self, state, scope, pos, paramTypes, a):
        if len(a) == 0:
            raise PFARuntimeException("empty array", self.errcodeBase + 0, self.name, pos)
        else:
            return argHighestN(a, 1, lambda x, y: compare(jsonNodeToAvroType(paramTypes[0]).items, x, y) < 0)[0]
provide(Argmax())

class Argmin(LibFcn):
    name = prefix + "argmin"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}], P.Int())
    errcodeBase = 15330
    def __call__(self, state, scope, pos, paramTypes, a):
        if len(a) == 0:
            raise PFARuntimeException("empty array", self.errcodeBase + 0, self.name, pos)
        else:
            return argLowestN(a, 1, lambda x, y: compare(jsonNodeToAvroType(paramTypes[0]).items, x, y) < 0)[0]
provide(Argmin())

class ArgmaxLT(LibFcn):
    name = prefix + "argmaxLT"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"lessThan": P.Fcn([P.Wildcard("A"), P.Wildcard("A")], P.Boolean())}], P.Int())
    errcodeBase = 15340
    def __call__(self, state, scope, pos, paramTypes, a, lessThan):
        if len(a) == 0:
            raise PFARuntimeException("empty array", self.errcodeBase + 0, self.name, pos)
        else:
            return argHighestN(a, 1, toLt(state, scope, lessThan))[0]
provide(ArgmaxLT())

class ArgminLT(LibFcn):
    name = prefix + "argminLT"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"lessThan": P.Fcn([P.Wildcard("A"), P.Wildcard("A")], P.Boolean())}], P.Int())
    errcodeBase = 15350
    def __call__(self, state, scope, pos, paramTypes, a, lessThan):
        if len(a) == 0:
            raise PFARuntimeException("empty array", self.errcodeBase + 0, self.name, pos)
        else:
            return argLowestN(a, 1, toLt(state, scope, lessThan))[0]
provide(ArgminLT())

class ArgmaxN(LibFcn):
    name = prefix + "argmaxN"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"n": P.Int()}], P.Array(P.Int()))
    errcodeBase = 15360
    def __call__(self, state, scope, pos, paramTypes, a, n):
        if len(a) == 0:
            raise PFARuntimeException("empty array", self.errcodeBase + 0, self.name, pos)
        elif n < 0:
            raise PFARuntimeException("n < 0", self.errcodeBase + 1, self.name, pos)
        else:
            return argHighestN(a, n, lambda x, y: compare(jsonNodeToAvroType(paramTypes[0]).items, x, y) < 0)
provide(ArgmaxN())

class ArgminN(LibFcn):
    name = prefix + "argminN"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"n": P.Int()}], P.Array(P.Int()))
    errcodeBase = 15370
    def __call__(self, state, scope, pos, paramTypes, a, n):
        if len(a) == 0:
            raise PFARuntimeException("empty array", self.errcodeBase + 0, self.name, pos)
        elif n < 0:
            raise PFARuntimeException("n < 0", self.errcodeBase + 1, self.name, pos)
        else:
            return argLowestN(a, n, lambda x, y: compare(jsonNodeToAvroType(paramTypes[0]).items, x, y) < 0)
provide(ArgminN())

class ArgmaxNLT(LibFcn):
    name = prefix + "argmaxNLT"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"n": P.Int()}, {"lessThan": P.Fcn([P.Wildcard("A"), P.Wildcard("A")], P.Boolean())}], P.Array(P.Int()))
    errcodeBase = 15380
    def __call__(self, state, scope, pos, paramTypes, a, n, lessThan):
        if len(a) == 0:
            raise PFARuntimeException("empty array", self.errcodeBase + 0, self.name, pos)
        elif n < 0:
            raise PFARuntimeException("n < 0", self.errcodeBase + 1, self.name, pos)
        else:
            return argHighestN(a, n, toLt(state, scope, lessThan))
provide(ArgmaxNLT())

class ArgminNLT(LibFcn):
    name = prefix + "argminNLT"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"n": P.Int()}, {"lessThan": P.Fcn([P.Wildcard("A"), P.Wildcard("A")], P.Boolean())}], P.Array(P.Int()))
    errcodeBase = 15390
    def __call__(self, state, scope, pos, paramTypes, a, n, lessThan):
        if len(a) == 0:
            raise PFARuntimeException("empty array", self.errcodeBase + 0, self.name, pos)
        elif n < 0:
            raise PFARuntimeException("n < 0", self.errcodeBase + 1, self.name, pos)
        else:
            return argLowestN(a, n, toLt(state, scope, lessThan))
provide(ArgminNLT())

#################################################################### numerical

class Sum(LibFcn):
    name = prefix + "sum"
    sig = Sig([{"a": P.Array(P.Wildcard("A", oneOf = anyNumber))}], P.Wildcard("A"))
    errcodeBase = 15400
    def __call__(self, state, scope, pos, paramTypes, a):
        out = sum(a)
        if paramTypes[-1] == "int":
            if INT_MIN_VALUE <= out <= INT_MAX_VALUE:
                return out
            else:
                raise PFARuntimeException("int overflow", self.errcodeBase + 0, self.name, pos)
        elif paramTypes[-1] == "long":
            if LONG_MIN_VALUE <= out <= LONG_MAX_VALUE:
                return out
            else:
                raise PFARuntimeException("long overflow", self.errcodeBase + 1, self.name, pos)
        else:
            return out
provide(Sum())

class Product(LibFcn):
    name = prefix + "product"
    sig = Sig([{"a": P.Array(P.Wildcard("A", oneOf = anyNumber))}], P.Wildcard("A"))
    errcodeBase = 15410
    def __call__(self, state, scope, pos, paramTypes, a):
        if paramTypes[-1] == "int":
            out = reduce(lambda a, b: a * b, a, 1)
            if INT_MIN_VALUE <= out <= INT_MAX_VALUE:
                return out
            else:
                raise PFARuntimeException("int overflow", self.errcodeBase + 0, self.name, pos)
        elif paramTypes[-1] == "long":
            out = reduce(lambda a, b: a * b, a, 1)
            if LONG_MIN_VALUE <= out <= LONG_MAX_VALUE:
                return out
            else:
                raise PFARuntimeException("long overflow", self.errcodeBase + 1, self.name, pos)
        else:
            return reduce(lambda a, b: a * b, a, 1.0)
provide(Product())

class Lnsum(LibFcn):
    name = prefix + "lnsum"
    sig = Sig([{"a": P.Array(P.Double())}], P.Double())
    errcodeBase = 15420
    def ln(self, x):
        if x < 0.0:
            return float("nan")
        elif x == 0.0:
            return float("-inf")
        else:
            return math.log(x)
    def __call__(self, state, scope, pos, paramTypes, a):
        return sum([self.ln(x) for x in a])
provide(Lnsum())

class Mean(LibFcn):
    name = prefix + "mean"
    sig = Sig([{"a": P.Array(P.Double())}], P.Double())
    errcodeBase = 15430
    def __call__(self, state, scope, pos, paramTypes, a):
        numer = sum(a)
        denom = float(len(a))
        if denom == 0.0:
            if numer > 0.0:
                return float("inf")
            elif numer < 0.0:
                return float("-inf")
            else:
                return float("nan")
        else:
            return numer / denom
provide(Mean())

class GeoMean(LibFcn):
    name = prefix + "geomean"
    sig = Sig([{"a": P.Array(P.Double())}], P.Double())
    errcodeBase = 15440
    def __call__(self, state, scope, pos, paramTypes, a):
        if len(a) == 0:
            return float("nan")
        else:
            return reduce(lambda a, b: a * b, a, 1.0)**(1.0/len(a))
provide(GeoMean())

class Median(LibFcn):
    name = prefix + "median"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}], P.Wildcard("A"))
    errcodeBase = 15450
    def __call__(self, state, scope, pos, paramTypes, a):
        if len(a) == 0:
            raise PFARuntimeException("empty array", self.errcodeBase + 0, self.name, pos)
        sa = sorted(a, key=cmp_to_key(lambda x, y: compare(jsonNodeToAvroType(paramTypes[0]).items, x, y)))
        half = len(sa) // 2
        dataType = paramTypes[-1]

        if len(sa) % 2:
            return sa[half]
        if (dataType is "float") or (dataType is "double"):
            return (sa[half - 1] + sa[half])/2.0
        else:
            return sa[half - 1]
provide(Median())

class Mode(LibFcn):
    name = prefix + "mode"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}], P.Wildcard("A"))
    errcodeBase = 15470
    def __call__(self, state, scope, pos, paramTypes, a):
        if len(a) == 0:
            raise PFARuntimeException("empty array", self.errcodeBase + 0, self.name, pos)
        else:
            counter = {}
            lookup = {}
            for x in a:
                xx = json.dumps(x)
                if xx not in counter:
                    counter[xx] = 0
                    lookup[xx] = x
                counter[xx] += 1
            bestn = 0
            for n in list(counter.values()):
                if n > bestn:
                    bestn = n
            besties = [lookup[xx] for xx in counter if counter[xx] == bestn]
            return Median()(state, scope, pos, paramTypes, besties)
provide(Mode())

class NTile(LibFcn):
    name = prefix + "ntile"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"p": P.Double()}], P.Wildcard("A"))
    errcodeBase = 15460
    def __call__(self, state, scope, pos, paramTypes, a, p):
        if len(a) == 0:
            raise PFARuntimeException("empty array", self.errcodeBase + 0, self.name, pos)
        if math.isnan(p):
            raise PFARuntimeException("p not a number", self.errcodeBase + 1, self.name, pos)
        if p <= 0.0:
            return lowestN(a, 1, lambda x, y: compare(jsonNodeToAvroType(paramTypes[0]).items, x, y) < 0)[0]
        if p >= 1.0:
            return highestN(a, 1, lambda x, y: compare(jsonNodeToAvroType(paramTypes[0]).items, x, y) < 0)[0]
        sa = sorted(a, key=cmp_to_key(lambda x, y: compare(jsonNodeToAvroType(paramTypes[0]).items, x, y)))
        k = (len(a) - 1.0)*p
        f = math.floor(k)
        dataType = paramTypes[-1]
        if (dataType is "float") or (dataType is "double"):
            c = math.ceil(k)
            if f == c:
                return sa[int(k)]
            d0 = sa[int(f)] * (c - k)
            d1 = sa[int(c)] * (k - f)
            return d0 + d1
        else:
            if len(sa) % 2 == 1:
                return sa[int(f)]
            else:
                return sa[int(k)]
provide(NTile())

#################################################################### set or set-like functions

class Distinct(LibFcn):
    name = prefix + "distinct"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}], P.Array(P.Wildcard("A")))
    errcodeBase = 15490
    def __call__(self, state, scope, pos, paramTypes, a):
        out = []
        for ai in a:
            if ai not in out:
                out.append(ai)
        return out
provide(Distinct())

class SetEq(LibFcn):
    name = prefix + "seteq"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"b": P.Array(P.Wildcard("A"))}], P.Boolean())
    errcodeBase = 15500
    def __call__(self, state, scope, pos, paramTypes, a, b):
        return all(ai in b for ai in a) and all(bi in a for bi in b)
provide(SetEq())

class Union(LibFcn):
    name = prefix + "union"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"b": P.Array(P.Wildcard("A"))}], P.Array(P.Wildcard("A")))
    errcodeBase = 15510
    def __call__(self, state, scope, pos, paramTypes, a, b):
        out = []
        for ai in a:
            if ai not in out:
                out.append(ai)
        for bi in b:
            if bi not in out:
                out.append(bi)
        return out
provide(Union())

class Intersection(LibFcn):
    name = prefix + "intersection"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"b": P.Array(P.Wildcard("A"))}], P.Array(P.Wildcard("A")))
    errcodeBase = 15520
    def __call__(self, state, scope, pos, paramTypes, a, b):
        out = []
        for ai in a:
            if ai in b:
                out.append(ai)
        for bi in b:
            if bi in a:
                out.append(bi)
        return out
provide(Intersection())

class Diff(LibFcn):
    name = prefix + "diff"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"b": P.Array(P.Wildcard("A"))}], P.Array(P.Wildcard("A")))
    errcodeBase = 15530
    def __call__(self, state, scope, pos, paramTypes, a, b):
        out = []
        for ai in a:
            if ai not in b:
                out.append(ai)
        return out
provide(Diff())

class SymDiff(LibFcn):
    name = prefix + "symdiff"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"b": P.Array(P.Wildcard("A"))}], P.Array(P.Wildcard("A")))
    errcodeBase = 15540
    def __call__(self, state, scope, pos, paramTypes, a, b):
        out = []
        for ai in a:
            if ai not in b:
                out.append(ai)
        for bi in b:
            if bi not in a:
                out.append(bi)
        return out
provide(SymDiff())

class Subset(LibFcn):
    name = prefix + "subset"
    sig = Sig([{"little": P.Array(P.Wildcard("A"))}, {"big": P.Array(P.Wildcard("A"))}], P.Boolean())
    errcodeBase = 15550
    def __call__(self, state, scope, pos, paramTypes, a, b):
        return all(ai in b for ai in a)
provide(Subset())

class Disjoint(LibFcn):
    name = prefix + "disjoint"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"b": P.Array(P.Wildcard("A"))}], P.Boolean())
    errcodeBase = 15560
    def __call__(self, state, scope, pos, paramTypes, a, b):
        return all(ai not in b for ai in a) and all(bi not in a for bi in b)
provide(Disjoint())

#################################################################### functional programming

class MapApply(LibFcn):
    name = prefix + "map"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"fcn": P.Fcn([P.Wildcard("A")], P.Wildcard("B"))}], P.Array(P.Wildcard("B")))
    errcodeBase = 15570
    def __call__(self, state, scope, pos, paramTypes, a, fcn):
        return [callfcn(state, scope, fcn, [x]) for x in a]
provide(MapApply())

class MapWithIndex(LibFcn):
    name = prefix + "mapWithIndex"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"fcn": P.Fcn([P.Int(), P.Wildcard("A")], P.Wildcard("B"))}], P.Array(P.Wildcard("B")))
    errcodeBase = 15580
    def __call__(self, state, scope, pos, paramTypes, a, fcn):
        return [callfcn(state, scope, fcn, [i, x]) for i, x in enumerate(a)]
provide(MapWithIndex())

class Filter(LibFcn):
    name = prefix + "filter"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"fcn": P.Fcn([P.Wildcard("A")], P.Boolean())}], P.Array(P.Wildcard("A")))
    errcodeBase = 15590
    def __call__(self, state, scope, pos, paramTypes, a, fcn):
        return [x for x in a if callfcn(state, scope, fcn, [x])]
provide(Filter())

class FilterWithIndex(LibFcn):
    name = prefix + "filterWithIndex"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"fcn": P.Fcn([P.Int(), P.Wildcard("A")], P.Boolean())}], P.Array(P.Wildcard("A")))
    errcodeBase = 15600
    def __call__(self, state, scope, pos, paramTypes, a, fcn):
        return [x for i, x in enumerate(a) if callfcn(state, scope, fcn, [i, x])]
provide(FilterWithIndex())

class FilterMap(LibFcn):
    name = prefix + "filterMap"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"fcn": P.Fcn([P.Wildcard("A")], P.Union([P.Wildcard("B"), P.Null()]))}], P.Array(P.Wildcard("B")))
    errcodeBase = 15610
    def __call__(self, state, scope, pos, paramTypes, a, fcn):
        typeNames = [jsonNodeToAvroType(t).name for t in paramTypes[1]["ret"] if t != "null"]
        out = []
        for x in a:
            y = callfcn(state, scope, fcn, [x])
            if y is not None:
                if isinstance(y, dict) and len(y) == 1 and list(y.keys())[0] in typeNames:
                    tag, value = list(y.items())[0]
                else:
                    value = y
                out.append(value)
        return out
provide(FilterMap())

class FilterMapWithIndex(LibFcn):
    name = prefix + "filterMapWithIndex"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"fcn": P.Fcn([P.Int(), P.Wildcard("A")], P.Union([P.Wildcard("B"), P.Null()]))}], P.Array(P.Wildcard("B")))
    errcodeBase = 15620
    def __call__(self, state, scope, pos, paramTypes, a, fcn):
        typeNames = [jsonNodeToAvroType(t).name for t in paramTypes[1]["ret"] if t != "null"]
        out = []
        for i, x in enumerate(a):
            y = callfcn(state, scope, fcn, [i, x])
            if y is not None:
                if isinstance(y, dict) and len(y) == 1 and list(y.keys())[0] in typeNames:
                    tag, value = list(y.items())[0]
                else:
                    value = y
                out.append(value)
        return out
provide(FilterMapWithIndex())

class FlatMap(LibFcn):
    name = prefix + "flatMap"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"fcn": P.Fcn([P.Wildcard("A")], P.Array(P.Wildcard("B")))}], P.Array(P.Wildcard("B")))
    errcodeBase = 15630
    def __call__(self, state, scope, pos, paramTypes, a, fcn):
        out = []
        for x in a:
            out.extend(callfcn(state, scope, fcn, [x]))
        return out
provide(FlatMap())

class FlatMapWithIndex(LibFcn):
    name = prefix + "flatMapWithIndex"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"fcn": P.Fcn([P.Int(), P.Wildcard("A")], P.Array(P.Wildcard("B")))}], P.Array(P.Wildcard("B")))
    errcodeBase = 15640
    def __call__(self, state, scope, pos, paramTypes, a, fcn):
        out = []
        for i, x in enumerate(a):
            out.extend(callfcn(state, scope, fcn, [i, x]))
        return out
provide(FlatMapWithIndex())

class ZipMap(LibFcn):
    name = prefix + "zipmap"
    sig = Sigs([Sig([{"a": P.Array(P.Wildcard("A"))}, {"b": P.Array(P.Wildcard("B"))}, {"fcn": P.Fcn([P.Wildcard("A"), P.Wildcard("B")], P.Wildcard("Z"))}], P.Array(P.Wildcard("Z"))),
                Sig([{"a": P.Array(P.Wildcard("A"))}, {"b": P.Array(P.Wildcard("B"))}, {"c": P.Array(P.Wildcard("C"))}, {"fcn": P.Fcn([P.Wildcard("A"), P.Wildcard("B"), P.Wildcard("C")], P.Wildcard("Z"))}], P.Array(P.Wildcard("Z"))),
                Sig([{"a": P.Array(P.Wildcard("A"))}, {"b": P.Array(P.Wildcard("B"))}, {"c": P.Array(P.Wildcard("C"))}, {"d": P.Array(P.Wildcard("D"))}, {"fcn": P.Fcn([P.Wildcard("A"), P.Wildcard("B"), P.Wildcard("C"), P.Wildcard("D")], P.Wildcard("Z"))}], P.Array(P.Wildcard("Z")))])
    errcodeBase = 15650
    def __call__(self, state, scope, pos, paramTypes, *args):
        if len(set(map(len, args[:-1]))) != 1:
            raise PFARuntimeException("misaligned arrays", self.errcodeBase + 0, self.name, pos)
        fcn = args[-1]
        out = []
        for abcd in zip(*args[:-1]):
            out.append(callfcn(state, scope, fcn, list(abcd)))
        return out
provide(ZipMap())

class ZipMapWithIndex(LibFcn):
    name = prefix + "zipmapWithIndex"
    sig = Sigs([Sig([{"a": P.Array(P.Wildcard("A"))}, {"b": P.Array(P.Wildcard("B"))}, {"fcn": P.Fcn([P.Int(), P.Wildcard("A"), P.Wildcard("B")], P.Wildcard("Z"))}], P.Array(P.Wildcard("Z"))),
                Sig([{"a": P.Array(P.Wildcard("A"))}, {"b": P.Array(P.Wildcard("B"))}, {"c": P.Array(P.Wildcard("C"))}, {"fcn": P.Fcn([P.Int(), P.Wildcard("A"), P.Wildcard("B"), P.Wildcard("C")], P.Wildcard("Z"))}], P.Array(P.Wildcard("Z"))),
                Sig([{"a": P.Array(P.Wildcard("A"))}, {"b": P.Array(P.Wildcard("B"))}, {"c": P.Array(P.Wildcard("C"))}, {"d": P.Array(P.Wildcard("D"))}, {"fcn": P.Fcn([P.Int(), P.Wildcard("A"), P.Wildcard("B"), P.Wildcard("C"), P.Wildcard("D")], P.Wildcard("Z"))}], P.Array(P.Wildcard("Z")))])
    errcodeBase = 15660
    def __call__(self, state, scope, pos, paramTypes, *args):
        if len(set(map(len, args[:-1]))) != 1:
            raise PFARuntimeException("misaligned arrays", self.errcodeBase + 0, self.name, pos)
        fcn = args[-1]
        out = []
        for i, abcd in enumerate(zip(*args[:-1])):
            out.append(callfcn(state, scope, fcn, [i] + list(abcd)))
        return out
provide(ZipMapWithIndex())

class Reduce(LibFcn):
    name = prefix + "reduce"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"fcn": P.Fcn([P.Wildcard("A"), P.Wildcard("A")], P.Wildcard("A"))}], P.Wildcard("A"))
    errcodeBase = 15670
    def __call__(self, state, scope, pos, paramTypes, a, fcn):
        if len(a) == 0:
            raise PFARuntimeException("empty array", self.errcodeBase + 0, self.name, pos)
        else:
            return reduce(lambda x, y: callfcn(state, scope, fcn, [x, y]), a)
provide(Reduce())

class ReduceRight(LibFcn):
    name = prefix + "reduceRight"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"fcn": P.Fcn([P.Wildcard("A"), P.Wildcard("A")], P.Wildcard("A"))}], P.Wildcard("A"))
    errcodeBase = 15680
    def __call__(self, state, scope, pos, paramTypes, a, fcn):
        if len(a) == 0:
            raise PFARuntimeException("empty array", self.errcodeBase + 0, self.name, pos)
        else:
            return reduce(lambda x, y: callfcn(state, scope, fcn, [y, x]), reversed(a))
provide(ReduceRight())

class Fold(LibFcn):
    name = prefix + "fold"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"zero": P.Wildcard("B")}, {"fcn": P.Fcn([P.Wildcard("B"), P.Wildcard("A")], P.Wildcard("B"))}], P.Wildcard("B"))
    errcodeBase = 15690
    def __call__(self, state, scope, pos, paramTypes, a, zero, fcn):
        return reduce(lambda x, y: callfcn(state, scope, fcn, [x, y]), a, zero)
provide(Fold())

class FoldRight(LibFcn):
    name = prefix + "foldRight"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"zero": P.Wildcard("B")}, {"fcn": P.Fcn([P.Wildcard("A"), P.Wildcard("B")], P.Wildcard("B"))}], P.Wildcard("B"))
    errcodeBase = 15700
    def __call__(self, state, scope, pos, paramTypes, a, zero, fcn):
        return reduce(lambda y, x: callfcn(state, scope, fcn, [x, y]), reversed(a), zero)
provide(FoldRight())

class TakeWhile(LibFcn):
    name = prefix + "takeWhile"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"fcn": P.Fcn([P.Wildcard("A")], P.Boolean())}], P.Array(P.Wildcard("A")))
    errcodeBase = 15710
    def __call__(self, state, scope, pos, paramTypes, a, fcn):
        out = []
        for x in a:
            if callfcn(state, scope, fcn, [x]):
                out.append(x)
            else:
                break
        return out
provide(TakeWhile())

class DropWhile(LibFcn):
    name = prefix + "dropWhile"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"fcn": P.Fcn([P.Wildcard("A")], P.Boolean())}], P.Array(P.Wildcard("A")))
    errcodeBase = 15720
    def __call__(self, state, scope, pos, paramTypes, a, fcn):
        i = 0
        for ai in a:
            if callfcn(state, scope, fcn, [ai]):
                i += 1
            else:
                break
        return a[i:]
provide(DropWhile())

#################################################################### functional tests

class Any(LibFcn):
    name = prefix + "any"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"fcn": P.Fcn([P.Wildcard("A")], P.Boolean())}], P.Boolean())
    errcodeBase = 15730
    def __call__(self, state, scope, pos, paramTypes, a, fcn):
        return any(callfcn(state, scope, fcn, [x]) for x in a)
provide(Any())

class All(LibFcn):
    name = prefix + "all"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"fcn": P.Fcn([P.Wildcard("A")], P.Boolean())}], P.Boolean())
    errcodeBase = 15740
    def __call__(self, state, scope, pos, paramTypes, a, fcn):
        return all(callfcn(state, scope, fcn, [x]) for x in a)
provide(All())

class Corresponds(LibFcn):
    name = prefix + "corresponds"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"b": P.Array(P.Wildcard("B"))}, {"fcn": P.Fcn([P.Wildcard("A"), P.Wildcard("B")], P.Boolean())}], P.Boolean())
    errcodeBase = 15750
    def __call__(self, state, scope, pos, paramTypes, a, b, fcn):
        return len(a) == len(b) and all(callfcn(state, scope, fcn, [x, y]) for x, y in zip(a, b))
provide(Corresponds())

class CorrespondsWithIndex(LibFcn):
    name = prefix + "correspondsWithIndex"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"b": P.Array(P.Wildcard("B"))}, {"fcn": P.Fcn([P.Int(), P.Wildcard("A"), P.Wildcard("B")], P.Boolean())}], P.Boolean())
    errcodeBase = 15760
    def __call__(self, state, scope, pos, paramTypes, a, b, fcn):
        return len(a) == len(b) and all(callfcn(state, scope, fcn, [i, x, y]) for i, (x, y) in enumerate(zip(a, b)))
provide(CorrespondsWithIndex())

#################################################################### restructuring

class SlidingWindow(LibFcn):
    name = prefix + "slidingWindow"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"size": P.Int()}, {"step": P.Int()}], P.Array(P.Array(P.Wildcard("A"))))
    errcodeBase = 15770
    def __call__(self, state, scope, pos, paramTypes, a, size, step):
        if size < 1:
            raise PFARuntimeException("size < 1", self.errcodeBase + 0, self.name, pos)
        elif step < 1:
            raise PFARuntimeException("step < 1", self.errcodeBase + 1, self.name, pos)
        out = []
        i = 0
        for start in range(0, len(a), step):
            chunk = a[start:(start + size)]
            if len(chunk) == size:
                out.append(chunk)
                i += 1
                if i % 1000 == 0:
                    state.checkTime()
        return out
provide(SlidingWindow())

class Combinations(LibFcn):
    name = prefix + "combinations"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"size": P.Int()}], P.Array(P.Array(P.Wildcard("A"))))
    errcodeBase = 15780
    def __call__(self, state, scope, pos, paramTypes, a, size):
        if size < 1:
            raise PFARuntimeException("size < 1", self.errcodeBase + 0, self.name, pos)
        out = []
        i = 0
        for combination in itertools.combinations(a, size):
            i += 1
            if i % 1000 == 0:
                state.checkTime()
            out.append(list(combination))
        return out
provide(Combinations())

class Permutations(LibFcn):
    name = prefix + "permutations"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}], P.Array(P.Array(P.Wildcard("A"))))
    errcodeBase = 15790
    def __call__(self, state, scope, pos, paramTypes, a):
        out = []
        i = 0
        for permutation in itertools.permutations(a):
            i += 1
            if i % 1000 == 0:
                state.checkTime()
            out.append(list(permutation))
        return out
provide(Permutations())

class Flatten(LibFcn):
    name = prefix + "flatten"
    sig = Sig([{"a": P.Array(P.Array(P.Wildcard("A")))}], P.Array(P.Wildcard("A")))
    errcodeBase = 15800
    def __call__(self, state, scope, pos, paramTypes, a):
        return titus.util.flatten(a)
provide(Flatten())

class GroupBy(LibFcn):
    name = prefix + "groupby"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}, {"fcn": P.Fcn([P.Wildcard("A")], P.String())}], P.Map(P.Array(P.Wildcard("A"))))
    errcodeBase = 15810
    def __call__(self, state, scope, pos, paramTypes, a, fcn):
        out = {}
        for x in a:
            key = callfcn(state, scope, fcn, [x])
            if key not in out:
                out[key] = []
            out[key].append(x)
        return out
provide(GroupBy())

class Logsumexp(LibFcn):
    name = prefix + "logsumexp"
    sig = Sig([{"a": P.Array(P.Double())}], P.Double())
    errcodeBase = 15480
    def __call__(self, state, scope, pos, paramTypes, datum):
        if len(datum) == 0:
            return float("nan")
        else:
            dmax = max(datum)
            res = 0.0
            for v in datum:
                res += math.exp(v - dmax)
            return math.log(res) + dmax
provide(Logsumexp())



#!/usr/bin/env python

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

from titus.fcn import Fcn
from titus.fcn import LibFcn
from titus.signature import Sig
from titus.signature import Sigs
from titus.signature import Lifespan
from titus.signature import PFAVersion
from titus.datatype import *
from titus.errors import *
from titus.lib.core import INT_MIN_VALUE, INT_MAX_VALUE, LONG_MIN_VALUE, LONG_MAX_VALUE
from titus.util import bytesToString
import titus.P as P

provides = {}
def provide(fcn):
    provides[fcn.name] = fcn

prefix = "rand."

########################################################### raw numbers of various types

class RandomInt(LibFcn):
    name = prefix + "int"
    sig = Sigs([Sig([], P.Int()),
                Sig([{"low": P.Int()}, {"high": P.Int()}], P.Int())])
    errcodeBase = 34000
    def __call__(self, state, scope, pos, paramTypes, *args):
        if len(args) == 0:
            return state.rand.randint(INT_MIN_VALUE, INT_MAX_VALUE)
        else:
            if args[1] <= args[0]: raise PFARuntimeException("high must be greater than low", self.errcodeBase + 0, self.name, pos)
            return state.rand.randint(args[0], args[1] - 1)  # Python's randint has an inclusive upper bound
provide(RandomInt())

class RandomLong(LibFcn):
    name = prefix + "long"
    sig = Sigs([Sig([], P.Long()),
                Sig([{"long": P.Long()}, {"high": P.Long()}], P.Long())])
    errcodeBase = 34010
    def __call__(self, state, scope, pos, paramTypes, *args):
        if len(args) == 0:
            return state.rand.randint(LONG_MIN_VALUE, LONG_MAX_VALUE)
        else:
            if args[1] <= args[0]: raise PFARuntimeException("high must be greater than low", self.errcodeBase + 0, self.name, pos)
            return state.rand.randint(args[0], args[1] - 1)  # Python's randint has an inclusive upper bound
provide(RandomLong())

class RandomFloat(LibFcn):
    name = prefix + "float"
    sig = Sig([{"low": P.Float()}, {"high": P.Float()}], P.Float())
    errcodeBase = 34020
    def __call__(self, state, scope, pos, paramTypes, low, high):
        if high <= low: raise PFARuntimeException("high must be greater than low", self.errcodeBase + 0, self.name, pos)
        return state.rand.uniform(low, high)
provide(RandomFloat())

class RandomDouble(LibFcn):
    name = prefix + "double"
    sig = Sig([{"low": P.Double()}, {"high": P.Double()}], P.Double())
    errcodeBase = 34030
    def __call__(self, state, scope, pos, paramTypes, low, high):
        if high <= low: raise PFARuntimeException("high must be greater than low", self.errcodeBase + 0, self.name, pos)
        return state.rand.uniform(low, high)
provide(RandomDouble())

########################################################### items from arrays

class RandomChoice(LibFcn):
    name = prefix + "choice"
    sig = Sig([{"population": P.Array(P.Wildcard("A"))}], P.Wildcard("A"))
    errcodeBase = 34040
    def __call__(self, state, scope, pos, paramTypes, population):
        if len(population) == 0:
            raise PFARuntimeException("population must not be empty", self.errcodeBase + 0, self.name, pos)
        return population[state.rand.randint(0, len(population) - 1)]
provide(RandomChoice())

class RandomChoices(LibFcn):
    name = prefix + "choices"
    sig = Sig([{"size": P.Int()}, {"population": P.Array(P.Wildcard("A"))}], P.Array(P.Wildcard("A")))
    errcodeBase = 34050
    def __call__(self, state, scope, pos, paramTypes, size, population):
        if len(population) == 0:
            raise PFARuntimeException("population must not be empty", self.errcodeBase + 0, self.name, pos)
        return [population[state.rand.randint(0, len(population) - 1)] for x in range(size)]
provide(RandomChoices())

class RandomSample(LibFcn):
    name = prefix + "sample"
    sig = Sig([{"size": P.Int()}, {"population": P.Array(P.Wildcard("A"))}], P.Array(P.Wildcard("A")))
    errcodeBase = 34060
    def __call__(self, state, scope, pos, paramTypes, size, population):
        if len(population) == 0:
            raise PFARuntimeException("population must not be empty", self.errcodeBase + 0, self.name, pos)
        if len(population) < size:
            raise PFARuntimeException("population smaller than requested subsample", self.errcodeBase + 1, self.name, pos)
        return state.rand.sample(population, size)
provide(RandomSample())
        
########################################################### deviates from a histogram

class RandomHistogram(LibFcn):
    name = prefix + "histogram"
    sig = Sigs([Sig([{"distribution": P.Array(P.Double())}], P.Int()),
                Sig([{"distribution": P.Array(P.WildRecord("A", {"prob": P.Double()}))}], P.Wildcard("A"))])
    errcodeBase = 34070
    def selectIndex(self, rand, distribution, pos):
        cumulativeSum = [0.0]
        for x in distribution:
            if math.isnan(x) or math.isinf(x):
                raise PFARuntimeException("distribution must be finite", self.errcodeBase + 1, self.name, pos)
            elif x < 0:
                raise PFARuntimeException("distribution must be non-negative", self.errcodeBase + 2, self.name, pos)
            cumulativeSum.append(cumulativeSum[-1] + x)
        total = cumulativeSum[-1]
        if total == 0.0:
            raise PFARuntimeException("distribution must be non-empty", self.errcodeBase + 0, self.name, pos)
        position = rand.uniform(0.0, total)
        for i, y in enumerate(cumulativeSum):
            if position < y:
                return i - 1
    def __call__(self, state, scope, pos, paramTypes, distribution):
        if isinstance(paramTypes[-1], dict) and paramTypes[-1].get("type") == "record":
            probs = [x["prob"] for x in distribution]
            index = self.selectIndex(state.rand, probs, pos)
            return distribution[index]
        else:
            return self.selectIndex(state.rand, distribution, pos)
provide(RandomHistogram())

########################################################### strings and byte arrays

class RandomString(LibFcn):
    name = prefix + "string"
    sig = Sigs([Sig([{"size": P.Int()}], P.String()),
                Sig([{"size": P.Int()}, {"population": P.String()}], P.String()),
                Sig([{"size": P.Int()}, {"low": P.Int()}, {"high": P.Int()}], P.String())])
    errcodeBase = 34080
    def __call__(self, state, scope, pos, paramTypes, size, *args):
        if size <= 0: raise PFARuntimeException("size must be positive", self.errcodeBase + 0, self.name, pos)
        if len(args) == 0:
            return "".join(chr(state.rand.randint(1, 0xD800)) for x in range(size))
        elif len(args) == 1:
            if len(args[0]) == 0:
                raise PFARuntimeException("population must be non-empty", self.errcodeBase + 3, self.name, pos)
            return "".join(args[0][state.rand.randint(0, len(args[0]) - 1)] for x in range(size))
        else:
            low, high = args
            if high <= low: raise PFARuntimeException("high must be greater than low", self.errcodeBase + 1, self.name, pos)
            if low < 1 or low > 0xD800 or high < 1 or high > 0xD800: raise PFARuntimeException("invalid char", self.errcodeBase + 2, self.name, pos)
            return "".join(chr(state.rand.randint(low, high - 1)) for x in range(size))
provide(RandomString())

class RandomBytes(LibFcn):
    name = prefix + "bytes"
    sig = Sigs([Sig([{"size": P.Int()}], P.Bytes()),
                Sig([{"size": P.Int()}, {"population": P.Bytes()}], P.Bytes()),
                Sig([{"size": P.Int()}, {"low": P.Int()}, {"high": P.Int()}], P.Bytes())])
    errcodeBase = 34090
    def __call__(self, state, scope, pos, paramTypes, size, *args):
        if size <= 0: raise PFARuntimeException("size must be positive", self.errcodeBase + 0, self.name, pos)
        if len(args) == 0:
            return "".join(chr(state.rand.randint(0, 255)) for x in range(size))
        elif len(args) == 1:
            if len(args[0]) == 0:
                raise PFARuntimeException("population must be non-empty", self.errcodeBase + 3, self.name, pos)
            if isinstance(args[0], str):
                population = list(args[0])
            elif isinstance(args[0], bytes):
                population = list(bytesToString(args[0]))
            return "".join(population[state.rand.randint(0, len(args[0]) - 1)] for x in range(size))
        else:
            low, high = args
            if high <= low: raise PFARuntimeException("high must be greater than low", self.errcodeBase + 1, self.name, pos)
            if low < 0 or low > 255 or high < 0 or high > 256: raise PFARuntimeException("invalid byte", self.errcodeBase + 2, self.name, pos)
            return "".join(chr(state.rand.randint(low, high - 1)) for x in range(size))
provide(RandomBytes())

class RandomUUID(LibFcn):
    name = prefix + "uuid"
    sig = Sig([], P.String(), Lifespan(None, PFAVersion(0, 7, 2), PFAVersion(0, 9, 0), "use rand.uuid4 instead"))
    errcodeBase = 34100
    def __call__(self, state, scope, pos, paramTypes):
        return "{0:08x}-{1:04x}-4{2:03x}-8{3:03x}-{4:016x}".format(state.rand.getrandbits(32), state.rand.getrandbits(16), state.rand.getrandbits(12), state.rand.getrandbits(12), state.rand.getrandbits(64))
provide(RandomUUID())

class RandomUUID4(LibFcn):
    name = prefix + "uuid4"
    sig = Sig([], P.String())
    errcodeBase = 34110
    def __call__(self, state, scope, pos, paramTypes):
        return "{0:08x}-{1:04x}-4{2:03x}-8{3:03x}-{4:016x}".format(state.rand.getrandbits(32), state.rand.getrandbits(16), state.rand.getrandbits(12), state.rand.getrandbits(12), state.rand.getrandbits(64))
provide(RandomUUID4())

########################################################### common probability distributions

class Gaussian(LibFcn):
    name = prefix + "gaussian"
    sig = Sig([{"mu": P.Double()}, {"sigma": P.Double()}], P.Double())
    errcodeBase = 34120
    def __call__(self, state, scope, pos, paramTypes, mu, sigma):
        return state.rand.gauss(mu, sigma)
provide(Gaussian())




#!/usr/bin/env python

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

import numpy

from titus.pfaast import Ast
from titus.pfaast import Call
from titus.pfaast import Ref
from titus.pfaast import LiteralNull
from titus.pfaast import LiteralBoolean
from titus.pfaast import LiteralInt
from titus.pfaast import LiteralLong
from titus.pfaast import LiteralFloat
from titus.pfaast import LiteralDouble
from titus.pfaast import LiteralString
from titus.datatype import AvroArray
from titus.datatype import AvroMap
from titus.datatype import AvroRecord
from titus.prettypfa import ppfa
from titus.prettypfa import pfa

class Transformation(object):
    """Represents an expression that can be applied to Numpy arrays in Python and included in a PFA scoring engine."""

    constants = {
        "m.pi": "math.pi",
        "m.e": "math.e",
        }

    functions = {
        "+": "numpy.add",
        "-": "numpy.subtract",
        "*": "numpy.multiply",
        "/": "numpy.true_divide",
        "u-": "numpy.negative",
        "**": "numpy.power",
        "%": "numpy.mod",
        "%%": "numpy.remainder",
        
        "==": "numpy.equal",
        "!=": "numpy.not_equal",
        ">": "numpy.greater",
        ">=": "numpy.greater_equal",
        "<": "numpy.less",
        "<=": "numpy.less_equal",
        "&&": "numpy.logical_and",
        "||": "numpy.logical_or",
        "^^": "numpy.logical_xor",
        "!": "numpy.logical_not",
        
        "m.abs": "numpy.absolute",
        "m.acos": "numpy.arccos",
        "m.asin": "numpy.arcsin",
        "m.atan": "numpy.arctan",
        "m.atan2": "numpy.arctan2",
        "m.ceil": "numpy.ceil",
        "m.copysign": "numpy.copysign",
        "m.cos": "numpy.cos",
        "m.cosh": "numpy.cosh",
        "m.exp": "numpy.exp",
        "m.expm1": "numpy.expm1",
        "m.floor": "numpy.floor",
        "m.hypot": "numpy.hypot",
        "m.ln": "numpy.log",
        "m.log10": "numpy.log10",
        "m.round": "numpy.round",
        "m.rint": "numpy.rint",
        "m.sin": "numpy.sin",
        "m.sinh": "numpy.sinh",
        "m.sqrt": "numpy.sqrt",
        "m.tan": "numpy.tan",
        "m.tanh": "numpy.tanh",
        }

    namespace = {"math": math, "numpy": numpy}

    @staticmethod
    def findFields(x):
        """Find the symbols (variables) referenced in a PFA expression.

        :type x: Pythonized JSON
        :param x: the PFA expression
        :rtype: list of strings
        :return: all symbols found
        """

        if isinstance(x, dict):
            return sum((Transformation.findFields(v) for k, v in list(x.items()) if k != "@"), [])
        elif isinstance(x, (list, tuple)):
            return sum((Transformation.findFields(xi) for xi in x), [])
        elif isinstance(x, str):
            return [x]
        else:
            return []

    @staticmethod
    def replace(x, subs):
        """Apply replacements to a PFA expression.

        :type x: Pythonized JSON
        :param x: the PFA expression
        :type subs: dict from substitution identifier to its replacement value
        :param subs: identifiers to replace
        :rtype: Pythonized JSON
        :return: the PFA with replacements
        """

        if isinstance(x, dict):
            return dict((k, Transformation.replace(v, subs)) for k, v in list(x.items()))
        elif isinstance(x, (list, tuple)):
            return [Transformation.replace(xi, subs) for xi in x]
        elif isinstance(x, str) and x in subs:
            return subs[x]
        else:
            return x

    @staticmethod
    def toNumpyExpr(ast):
        """Convert a PFA abstract syntax tree into a Numpy expression.

        :type ast: titus.pfaast.Ast
        :param ast: the PFA to convert
        :rtype: string
        :return: executable Python code in a string
        """

        if isinstance(ast, Call):
            if ast.name in Transformation.constants and len(ast.args) == 0:
                return Transformation.constants[ast.name]
            elif ast.name in Transformation.functions:
                return Transformation.functions[ast.name] + "(" + ", ".join(Transformation.toNumpyExpr(x) for x in ast.args) + ")"
            else:
                raise ValueError("No numpy equivalent defined for function {0}".format(ast.name))
        elif isinstance(ast, Ref):
            return ast.name
        elif isinstance(ast, LiteralNull):
            return float("nan")
        elif isinstance(ast, (LiteralBoolean, LiteralInt, LiteralLong, LiteralFloat, LiteralDouble, LiteralString)):
            return repr(ast.value)
        else:
            raise ValueError("No numpy equivalent defined for expression {0}".format(ast.toJson()))

    def __init__(self, *indexed, **named):
        """Create a Transformation either from an ordered list of Python expressions or by keywords.

        Although it's possible to mix the ordered list method and the keyword method, doing so could be confusing.

        Expressions loaded as positional arguments are given keyword names "_0", "_1", "_2", etc., and all methods proceed as though they were loaded as keywords.

        :type indexed: strings (as positional arguments)
        :param indexed: executable Python code in strings
        :type named: strings (as keyword arguments)
        :param named: executable Python code in strings
        """

        # indexed inputs come first, in order, and are labeled as _0, _1, _2, etc.
        self.exprs = named
        self.order = sorted(named.keys())
        for i, expr in enumerate(indexed):
            key = "_%d" % i
            self.exprs[key] = expr
            self.order.insert(i, key)

        # construct PFA from each expression
        self.pfas = dict((k, ppfa(v).jsonNode(False, set())) for k, v in list(self.exprs.items()))

        # find all the fields referenced in all of the expressions
        self.fields = sorted(set(sum((Transformation.findFields(x) for x in list(self.pfas.values())), [])))

        # construct lambda functions for transforming Numpy
        self.lambdas = dict((k, eval("lambda " + ", ".join(self.fields) + ": " + Transformation.toNumpyExpr(ppfa(v)), self.namespace)) for k, v in list(self.exprs.items()))

    def transform(self, dataset, fieldNames=None):
        """Return a transformed Numpy dataset (leaving the original intact).

        :type dataset: Numpy record array, dict of 1-D Numpy arrays, or Numpy 2-d table
        :param dataset: input dataset to be transformed; the Numpy record names or dict keys must correspond to the keywords of the arguments used to construct this ``Transformation``, or the column indexes of the 2-d table must correspond to the positions of the arguments used to construct this ``Transformation``.
        :rtype: same as ``dataset``
        :return: transformed dataset (operations are *not* performed in-place)
        """

        # option 1: dataset is a record array
        if isinstance(dataset, numpy.core.records.recarray):
            outType = "recarray"
            if fieldNames is None:
                fieldNames = dataset.dtype.names

        # option 2: dataset is a dictionary of 1-D arrays
        elif isinstance(dataset, dict) and all(isinstance(x, numpy.ndarray) and len(x.shape) == 1 for x in list(dataset.values())):
            outType = "dict"
            if fieldNames is None:
                fieldNames = list(dataset.keys())

        # option 3: dataset is a 2-D table
        elif isinstance(dataset, numpy.ndarray):
            outType = "array"
            if fieldNames is None:
                fieldNames = self.fields
            if len(dataset.shape) == 1:
                dataset = numpy.reshape(dataset, (dataset.shape[0], 1))
            if len(fieldNames) != dataset.shape[1]:
                raise TypeError("cannot interpret a {0}-column array as {1} fields".format(dataset.shape[1], len(fieldNames)))

        else:
            raise TypeError("expecting Numpy recarray, dictionary of 1-D arrays, or a 2-D array")

        # complain if any fields are not in this fieldNames
        cannotSupply = set(self.fields).difference(set(fieldNames))
        if len(cannotSupply) > 0:
            raise TypeError("expressions need [{0}], which are not supplied".format(", ".join(sorted(cannotSupply))))

        # evaluate the Numpy expressions
        if outType == "array":
            computed = dict((k, v(*[dataset[:, fieldNames.index(f)] for f in self.fields])) for k, v in list(self.lambdas.items()))
        else:
            computed = dict((k, v(*[dataset[f] for f in self.fields])) for k, v in list(self.lambdas.items()))

        # return the same type you received
        if outType == "recarray":
            return numpy.core.records.fromarrays([computed[x] for x in self.order], names=self.order)
        elif outType == "dict":
            return computed
        elif outType == "array":
            return numpy.vstack([computed[k] for k in self.order]).T

    @staticmethod
    def interpret(x):
        """Interpret expression from a PFA abstract syntax tree or PrettyPFA string.

        :type x: titus.datatype.Ast, PrettyPFA string
        :param x: input PFA
        :rtype: Pythonized JSON
        :return: the PFA as Pythonized JSON
        """

        if isinstance(x, Ast):
            return x.jsonNode(False, set())
        elif isinstance(x, str):
            return ppfa(x).jsonNode(False, set())
        else:
            return pfa(x).jsonNode(False, set())

    def new(self, avroType, subs={}, **subs2):
        """Construct a PFA "new" expression for this transformation.

        :type avroType: titus.datatype.AvroType
        :param avroType: type argument for the PFA "new" expression
        :type subs: dict of substitutions
        :param subs: subsititusions to apply
        :type subs2: dict of substitutions
        :param subs2: more subsititusions to apply
        :rtype: Pythonized JSON
        :return: PFA "new" expression
        """

        subs2.update(subs)
        for k, v in list(subs2.items()):
            subs2[k] = Transformation.interpret(v)

        if isinstance(avroType, AvroArray):
            return {"type": avroType.jsonNode(set()),
                    "new": [Transformation.replace(self.pfas[k], subs2) for k in self.order]}
        elif isinstance(avroType, dict) and avroType["type"] == "array":
            return {"type": avroType,
                    "new": [Transformation.replace(self.pfas[k], subs2) for k in self.order]}

        elif isinstance(avroType, AvroMap):
            return {"type": avroType.jsonNode(set()),
                    "new": dict((k, Transformation.replace(v, subs2)) for k, v in list(self.pfas.items()))}
        elif isinstance(avroType, dict) and avroType["type"] == "map":
            return {"type": avroType,
                    "new": dict((k, Transformation.replace(v, subs2)) for k, v in list(self.pfas.items()))}

        elif isinstance(avroType, AvroRecord):
            return {"type": avroType.name,
                    "new": dict((k, Transformation.replace(v, subs2)) for k, v in list(self.pfas.items()))}
        elif isinstance(avroType, dict) and avroType["type"] == "record" or isinstance(avroType, str):
            return {"type": avroType,
                    "new": dict((k, Transformation.replace(v, subs2)) for k, v in list(self.pfas.items()))}

        else:
            raise TypeError("new can only be used to make an array, map, or record")

    def let(self, subs={}, **subs2):
        """Construct a PFA "let" expression for this transformation.

        :type subs: dict of substitutions
        :param subs: subsititusions to apply
        :type subs2: dict of substitutions
        :param subs2: more subsititusions to apply
        :rtype: Pythonized JSON
        :return: PFA "let" expression
        """

        subs2.update(subs)
        for k, v in list(subs2.items()):
            subs2[k] = Transformation.interpret(v)

        return {"let": dict((k, Transformation.replace(v, subs2)) for k, v in list(self.pfas.items()))}

    def expr(self, name="_0", subs={}, **subs2):
        """Construct a PFA expression for one of the expressions in this ``Transformation``.

        :type name: string
        :param name: expression to choose
        :type subs: dict of substitutions
        :param subs: subsititusions to apply
        :type subs2: dict of substitutions
        :param subs2: more subsititusions to apply
        :rtype: Pythonized JSON
        :return: PFA "let" expression
        """

        subs2.update(subs)
        for k, v in list(subs2.items()):
            subs2[k] = Transformation.interpret(v)

        return Transformation.replace(self.pfas[name], subs2)




#!/usr/bin/env python

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import random
from collections import OrderedDict
from functools import cmp_to_key

import numpy

from titus.producer.transformation import Transformation
import titus.prettypfa

### metrics are nested objects like Euclidean(AbsDiff()), and they may
### be user-defined.  They mirror the metrics and similarity functions
### available in PFA.

### interfaces

def _NotImplementedError():
    raise NotImplementedError
class Similarity(object):
    """Trait for similarity functions in Numpy and PFA (compare two scalars, return a non-negative number)."""
    def __init__(self):
        self.calculate = lambda dataset, cluster: _NotImplementedError()
    def pfa(self):
        raise NotImplementedError
class Metric(object):
    """Trait for metric functions in Numpy and PFA (compare two vectors, return a non-negative number)."""
    def __init__(self):
        self.calculate = lambda dataset, cluster: _NotImplementedError()
    def pfa(self):
        raise NotImplementedError

### similarity

class AbsDiff(Similarity):
    """Absolute difference similarity function for Numpy and PFA."""
    def __init__(self):
        self.calculate = lambda dataset, cluster: numpy.absolute(dataset - cluster)
    def pfa(self):
        return {"fcn": "metric.absDiff"}

class GaussianSimilarity(Similarity):
    """Gaussian similarity function for Numpy and PFA."""
    def __init__(self, sigma):
        self.calculate = lambda dataset, cluster: numpy.exp(-numpy.log(2) * numpy.square(dataset - cluster) / sigma**2)
        self.sigma = sigma
    def pfa(self):
        x = "similarityX"
        y = "similarityY"
        return {"params": [{x: "double"}, {y: "double"}],
                "ret": "double",
                "do": {"metric.gaussianSimilarity": [x, y, self.sigma]}}

### metrics

class Euclidean(Metric):
    """Euclidean metric for Numpy and PFA."""
    def __init__(self, similarity):
        self.calculate = lambda dataset, cluster: numpy.sqrt(numpy.sum(numpy.square(similarity.calculate(dataset, cluster)), axis=1))
        self.similarity = similarity
    def pfa(self, x, y):
        return {"metric.euclidean": [self.similarity.pfa(), x, y]}

class SquaredEuclidean(Metric):
    """Squared euclidean metric for Numpy and PFA."""
    def __init__(self, similarity):
        self.calculate = lambda dataset, cluster: numpy.sum(numpy.square(similarity.calculate(dataset, cluster)), axis=1)
        self.similarity = similarity
    def pfa(self, x, y):
        x = "metricX"
        y = "metricY"
        return {"metric.squaredEuclidean": [self.similarity.pfa(), x, y]}

class Chebyshev(Metric):
    """Chebyshev (maximum) metric for Numpy and PFA."""
    def __init__(self, similarity):
        self.calculate = lambda dataset, cluster: numpy.max(similarity.calculate(dataset, cluster), axis=1)
        self.similarity = similarity
    def pfa(self, x, y):
        x = "metricX"
        y = "metricY"
        return {"metric.chebyshev": [self.similarity.pfa(), x, y]}

class Taxicab(Metric):
    """Taxicab (sum) metric for Numpy and PFA."""
    def __init__(self, similarity):
        self.calculate = lambda dataset, cluster: numpy.sum(similarity.calculate(dataset, cluster), axis=1)
        self.similarity = similarity
    def pfa(self, x, y):
        x = "metricX"
        y = "metricY"
        return {"metric.taxicab": [self.similarity.pfa(), x, y]}

class Minkowski(Metric):
    """Minkowski metric for Numpy and PFA."""
    def __init__(self, similarity, p):
        self.calculate = lambda dataset, cluster: numpy.pow(numpy.sum(numpy.pow(similarity.calculate(dataset, cluster), p), axis=1), 1.0/p)
        self.similarity = similarity
        self.p = p
    def pfa(self, x, y):
        x = "metricX"
        y = "metricY"
        return {"metric.minkowski": [self.similarity.pfa(), x, y, self.p]}

### stopping conditions are functions that take iterationNumber (int),
# corrections (Python list of Numpy arrays), datasetSize (int) and
# return bool (continue iterating if True)

### they may be user-defined or constructed from these functions like
# whileall(printChange("6.4f"), halfChange(0.001), clusterJumped()) to
# print iteration data, stop when at least half change by less than
# 0.001, and keep going if one jumped

def printValue(format="g"):
    """Generates a "stopping condition" that prints the current value and never stops.

    :type format: string
    :param format: format string ("g" is general number, "8.3f" is 8-characters wide, 3-digits after the decimal floating point, etc.)
    :rtype: callable that takes iterationNumber, corrections, values, datasetSize as arguments
    :return: stopping condition function
    """

    state = {"ready": False}
    def out(iterationNumber, corrections, values, datasetSize):
        if not state["ready"]:
            state["j"] = "{:5s} (jump)"
            for v in values:
                if v is not None:
                    state["n"] = "{0:5s}" + "".join((" {%d:%s}" % (i + 1, format)) for i in range(len(v)))
                    break
            if "n" in state:
                state["ready"] = True
            print("iter  values")
            print("----------------------------------")
        for index, v in enumerate(values):
            if index == 0:
                it = repr(iterationNumber)
            else:
                it = ""
            if v is None:
                print(state["j"].format(it))
            else:
                print(state["n"].format(it, *v))
        return True
    return out

def printChange(format="g"):
    """Generates a "stopping condition" that prints changes in values and never stops.

    :type format: string
    :param format: format string ("g" is general number, "8.3f" is 8-characters wide, 3-digits after the decimal floating point, etc.)
    :rtype: callable that takes iterationNumber, corrections, values, datasetSize as arguments
    :return: stopping condition function
    """

    state = {"ready": False}
    def out(iterationNumber, corrections, values, datasetSize):
        if not state["ready"]:
            state["j"] = "{:5s} (jump)"
            for corr in corrections:
                if corr is not None:
                    state["n"] = "{0:5s}" + "".join((" {%d:%s}" % (i + 1, format)) for i in range(len(corr)))
                    break
            if "n" in state:
                state["ready"] = True
            print("iter  changes")
            print("----------------------------------")
        for index, corr in enumerate(corrections):
            if index == 0:
                it = repr(iterationNumber)
            else:
                it = ""
            if corr is None:
                print(state["j"].format(it))
            else:
                print(state["n"].format(it, *corr))
        return True
    return out

def clusterJumped():
    """Generates a stopping condition that stops if no clusters jumped (reset to a random point because of encounting nan).

    :rtype: callable that takes iterationNumber, corrections, values, datasetSize as arguments
    :return: stopping condition function
    """
    return lambda iterationNumber, corrections, values, datasetSize: all(x is not None for x in corrections)

def maxIterations(number):
    """Generates a stopping condition that stops after a given number of iterations have passed.

    :rtype: callable that takes iterationNumber, corrections, values, datasetSize as arguments
    :return: stopping condition function
    """
    return lambda iterationNumber, corrections, values, datasetSize: iterationNumber < number

def allChange(threshold):
    """Generates a stopping condition that stops if all cluster changes are less than a threshold.

    :type threshold: number
    :param threshold: maximum change allowed for all clusters
    :rtype: callable that takes iterationNumber, corrections, values, datasetSize as arguments
    :return: stopping condition function
    """
    return lambda iterationNumber, corrections, values, datasetSize: not all((numpy.absolute(x) < threshold).all() for x in corrections if x is not None)

def halfChange(threshold):
    """Generates a stopping condition that stops if half of the cluster changes are less than a threshold.

    :type threshold: number
    :param threshold: maximum change allowed for half of the clusters
    :rtype: callable that takes iterationNumber, corrections, values, datasetSize as arguments
    :return: stopping condition function
    """
    return lambda iterationNumber, corrections, values, datasetSize: numpy.sum([(numpy.absolute(x) < threshold).all() for x in corrections if x is not None], dtype=numpy.dtype(float)) / numpy.sum([x is not None for x in corrections], dtype=numpy.dtype(float)) < 0.5

def whileall(*conditions):
    """Generates a stopping condition that continues while all of its subconditions continue.

    :type conditions: stopping condition functions
    :param conditions: subconditions
    :rtype: callable that takes iterationNumber, corrections, values, datasetSize as arguments
    :return: stopping condition function
    """
    return lambda *args: all(x(*args) for x in conditions)

def whileany(*conditions):
    """Generates a stopping condition that continues while any of its subconditions continue.

    :type conditions: stopping condition functions
    :param conditions: subconditions
    :rtype: callable that takes iterationNumber, corrections, values, datasetSize as arguments
    :return: stopping condition function
    """
    return lambda *args: any(x(*args) for x in conditions)

def moving():
    """Generates a stopping condition that stops when all clusters change less than 1e-15 and none are jumping (reset to a random point because of encounting nan).

    :type conditions: stopping condition functions
    :param conditions: subconditions
    :rtype: callable that takes iterationNumber, corrections, values, datasetSize as arguments
    :return: stopping condition function
    """
    return whileall(clusterJumped(), allChange(1e-15))

### the KMeans class

class KMeans(object):
    """Represents a k-means optimization by storing a dataset and performing all operations *in-place*.

    Usually, you would construct the object, possibly stepup, then optimize and export to pfaDocument.
    """

    def __init__(self, numberOfClusters, dataset, weights=None, metric=Euclidean(AbsDiff()), minPointsInCluster=None, maxPointsForClustering=None):
        """Construct a KMeans object, initializing cluster centers to unique, random points from the dataset.

        :type numberOfClusters: positive integer
        :param numberOfClusters: number of clusters (the "k" in k-means)
        :type dataset: 2-d Numpy array
        :param dataset: dataset to cluster; ``dataset.shape[0]`` is the number of records (rows), ``dataset.shape[1]`` is the number of dimensions for each point (columns)
        :type weights: 1-d Numpy array or ``None``
        :param weights: how much to weight each point in the ``dataset``: must have shape equal to ``(dataset.shape[0],)``; ``0`` means ignore the dataset, ``1`` means normal weight; ``None`` generates all ones
        :type metric: titus.produce.kmeans.Metric
        :param metric: metric for Numpy and PFA, such as ``Euclidean(AbsDiff())``
        :type minPointsInCluster: non-negative integer or ``None``
        :param minPointsInCluster: minimum number of points before jumping (replacing cluster with a random point during optimization)
        :type maxPointsForClustering: positive integer or ``None``
        :param maxPointsForClustering: maximum number of points in an optimization (if ``dataset.shape[0]`` exceeds this amount, a random subset is chosen)
        """

        if len(dataset.shape) != 2:
            raise TypeError("dataset must be two-dimensional: dataset.shape[0] is the number of records (rows), dataset.shape[1] is the number of dimensions (columns)")

        self.dataset = dataset
        if weights is not None and weights.shape != (dataset.shape[0],):
            raise TypeError("weights must have as many records as the dataset and must be one dimensional")
        self.weights = weights
        self.metric = metric

        try:
            flattenedView = numpy.ascontiguousarray(self.dataset).view(numpy.dtype((numpy.void, self.dataset.dtype.itemsize * self.dataset.shape[1])))
            _, indexes = numpy.unique(flattenedView, return_index=True)
            self.uniques = self.dataset[indexes]
        except TypeError:
            self.uniques = self.dataset

        if self.uniques.shape[0] <= numberOfClusters:
            raise TypeError("the number of unique records in the dataset ({0} in this case) must be strictly greater than numberOfClusters ({1})".format(self.uniques.shape[0], numberOfClusters))
        self.numberOfClusters = numberOfClusters

        self.clusters = []
        for index in range(numberOfClusters):
            self.clusters.append(self.newCluster())

        self.minPointsInCluster = minPointsInCluster
        self.maxPointsForClustering = maxPointsForClustering

    def randomPoint(self):
        """Pick a random point from the dataset.

        :rtype: 1-d Numpy array
        :return: a *copy* of a random point
        """
        # make sure to copy it, so there are no hidden connections between dataset and clusters
        return self.uniques[random.randint(0, self.uniques.shape[0] - 1),:].copy()

    def newCluster(self):
        """Pick a random point from the dataset and ensure that it is different from all other cluster centers.

        :rtype: 1-d Numpy array
        :return: a *copy* of a random point, guaranteed to be different from all other clusters.
        """
        newCluster = self.randomPoint()
        while any(numpy.array_equal(x, newCluster) for x in self.clusters):
            newCluster = self.randomPoint()
        return newCluster

    def randomSubset(self, subsetSize):
        """Return a (dataset, weights) that are randomly chosen to have ``subsetSize`` records.

        :type subsetSize: positive integer
        :param subsetSize: size of the sample
        :rtype: (2-d Numpy array, 1-d Numpy array)
        :return: (dataset, weights) sampled without replacement (if the original dataset is unique, the new one will be, too)
        """

        if subsetSize <= self.numberOfClusters:
            raise TypeError("subsetSize must be strictly greater than the numberOfClusters")

        indexes = random.sample(range(self.dataset.shape[0]), subsetSize)
        dataset = self.dataset[indexes,:]
        if self.weights is None:
            weights = None
        else:
            weights = self.weights[indexes]

        return dataset, weights

    def closestCluster(self, dataset=None, weights=None):
        """Identify the closest cluster to each element in the dataset.

        :type dataset: 2-d Numpy array or ``None``
        :param dataset: an input dataset or the built-in dataset if ``None`` is passed
        :type weights: 1-d Numpy array or ``None``
        :param weights: input weights or the built-in weights if ``None`` is passed
        :rtype: 1-d Numpy array of integers
        :return: the *indexes* of the closest cluster for each datum
        """

        if dataset is None:
            dataset = self.dataset
        if weights is None:
            weights = self.weights

        # distanceToCenter is the result of applying the metric to each point in the dataset, for each cluster
        # distanceToCenter.shape[0] is the number of records, distanceToCenter.shape[1] is the number of clusters

        distanceToCenter = numpy.empty((dataset.shape[0], self.numberOfClusters), dtype=numpy.dtype(float))
        for clusterIndex, cluster in enumerate(self.clusters):
            distanceToCenter[:, clusterIndex] = self.metric.calculate(dataset, cluster)

        # indexOfClosestCluster is the cluster classification for each point in the dataset
        return numpy.argmin(distanceToCenter, axis=1)

    def iterate(self, dataset, weights, iterationNumber, condition):
        """Perform one iteration step (in-place; modifies ``self.clusters``).

        :type dataset: 2-d Numpy array
        :param dataset: an input dataset
        :type weights: 1-d Numpy array
        :param weights: input weights
        :type iterationNumber: non-negative integer
        :param iterationNumber: the iteration number
        :type condition: callable that takes iterationNumber, corrections, values, datasetSize as arguments
        :param condition: the stopping condition
        :rtype: bool
        :return: the result of the stopping condition
        """

        indexOfClosestCluster = self.closestCluster(dataset, weights)

        values = []
        corrections = []
        for clusterIndex, cluster in enumerate(self.clusters):
            # select by cluster classification
            selection = (indexOfClosestCluster == clusterIndex)
            residuals = (dataset - cluster)[selection]

            # weights scale the residuals
            if weights is not None:
                residuals *= weights[selection]

            if self.minPointsInCluster is not None and numpy.sum(selection) < self.minPointsInCluster:
                # too few points in this cluster; jump to a new random point
                self.clusters[clusterIndex] = self.newCluster()
                values.append(None)
                corrections.append(None)

            else:
                # compute the mean of the displacements of points associated with this cluster
                # (note that the similarity metric used here is the trivial one, possibly different from the classification metric)
                correction = residuals.mean(axis=0)
                numpy.add(cluster, correction, cluster)

                if not numpy.isfinite(cluster).all():
                    # mean of a component is NaN or Inf, possibly because of zero points in the cluster
                    self.clusters[clusterIndex] = self.newCluster()
                    values.append(None)
                    corrections.append(None)
                else:
                    # good step: correction is not None
                    values.append(cluster)
                    corrections.append(correction)

        # call user-supplied test for continuation
        return condition(iterationNumber, corrections, values, dataset.shape[0])

    def stepup(self, condition, base=2):
        """Optimize the cluster set in successively larger subsets of the dataset.  (This can be viewed as a cluster seeding technique.)

        If randomly seeded, optimizing the whole dataset can be slow to converge: a long time per iteration times many iterations.

        Optimizing a random subset takes as many iterations, but the time per iteration is short.  However, the final cluster centers are only approximate.

        Optimizing the whole dataset with approximate cluster starting points takes a long time per iteration but fewer iterations.

        This procedure runs the k-means optimization technique on random subsets with exponentially increasing sizes from the smallest base**x that is larger than minPointsInCluster (or numberOfClusters) to the largest base**x that is a subset of the whole dataset.

        :type condition: callable that takes iterationNumber, corrections, values, datasetSize as arguments
        :param condition: the stopping condition
        :type base: integer greater than 1
        :param base: the factor by which the subset size is increased after each convergence
        :rtype: ``None``
        :return: nothing; modifies cluster set in-place
        """

        if self.minPointsInCluster is None:
            minPointsInCluster = self.numberOfClusters
        else:
            minPointsInCluster = max(self.numberOfClusters, self.minPointsInCluster)

        if self.maxPointsForClustering is None:
            maxPointsForClustering = self.dataset.shape[0]
        else:
            maxPointsForClustering = self.maxPointsForClustering

        trialSizes = [base**x for x in range(int(math.log(maxPointsForClustering, base)) + 1) if base**x > minPointsInCluster]

        for trialSize in trialSizes:
            dataset, weights = self.randomSubset(trialSize)
            
            iterationNumber = 0
            while self.iterate(dataset, weights, iterationNumber, condition):
                iterationNumber += 1

    def optimize(self, condition):
        """Run a standard k-means (Lloyd's algorithm) on the dataset, changing the clusters *in-place*.

        :type condition: callable that takes iterationNumber, corrections, values, datasetSize as arguments
        :param condition: the stopping condition
        :rtype: ``None``
        :return: nothing; modifies cluster set in-place
        """

        if self.maxPointsForClustering is None:
            dataset, weights = self.dataset, self.weights
        else:
            dataset, weights = self.randomSubset(self.maxPointsForClustering)

        iterationNumber = 0
        while self.iterate(dataset, weights, iterationNumber, condition):
            iterationNumber += 1

    def centers(self, sort=True):
        """Get the cluster centers as a sorted Python list (canonical form).

        :type sort: bool
        :param sort: if ``True``, sort the centers for stable results
        :rtype: list of list of numbers
        :return: the cluster centers as Pythonized JSON
        """
        if sort:
            centers = sorted(map(list, self.clusters))
        else:
            centers = list(map(list, self.clusters))
        return centers

    def pfaType(self, clusterTypeName, idType="string", centerComponentType="double", populations=False):
        """Create a PFA type schema representing this cluster set.

        :type clusterTypeName: string
        :param clusterTypeName: name of the PFA record type
        :type idType: Pythonized JSON
        :param idType: subtype for the ``id`` field
        :type centerComponentType: Pythonized JSON
        :param centerComponentType: subtype for the center array items
        :type populations: bool
        :param populations: if ``True``, include the number of training points as a "population" field
        :rtype: Pythonized JSON
        :return: PFA type schema for an array of clusters
        """

        fields = [{"name": "center", "type": {"type": "array", "items": centerComponentType}},
                  {"name": "id", "type": idType}]
        if populations:
            fields.append({"name": "population", "type": "int"})

        return {"type": "array",
                "items": {"type": "record",
                          "name": clusterTypeName,
                          "fields": fields}}
            
    def pfaValue(self, ids, populations=False, sort=True):
        """Create a PFA data structure representing this cluster set.

        :type ids: list of string
        :param ids: names of the clusters
        :type populations: bool
        :param populations: if ``True``, include the number of training points as a "population" field
        :type sort: bool
        :param sort: if ``True``, sort the centers for stable results
        :rtype: Pythonized JSON
        :return: data structure that should be inserted in the ``init`` section of the cell or pool containing the clusters
        """

        if len(ids) != self.numberOfClusters:
            raise TypeError("ids should be a list with length equal to the number of clusters")

        out = [{"center": x} for x in self.centers(sort=False)]

        if populations:
            indexOfClosestCluster = self.closestCluster(self.dataset, self.weights)
            for clusterIndex in range(len(self.clusters)):
                out[clusterIndex]["population"] = int(numpy.sum(indexOfClosestCluster == clusterIndex))

        if sort:
            indexes = [i for x, i in sorted(list(zip(map(list, self.clusters), range(len(self.clusters)))))]  
            out = list(numpy.array(out, dtype=object)[indexes])

        for idi, cluster in zip(ids, out):
            cluster["id"] = idi

        return out

    def pfaDocument(self, clusterTypeName, ids, populations=False, sort=True, preprocess=None, idType="string", dataComponentType="double", centerComponentType="double"):
        """Create a PFA document to score with this cluster set.

        :type clusterTypeName: string
        :param clusterTypeName: name of the PFA record type
        :type ids: list of string
        :param ids: names of the clusters
        :type populations: bool
        :param populations: if ``True``, include the number of training points as a "population" field
        :type sort: bool
        :param sort: if ``True``, sort the centers for stable results
        :type preprocess: PrettyPFA substitution or ``None``
        :param preprocess: pre-processing expression
        :type idType: Pythonized JSON
        :param idType: subtype for the ``id`` field
        :type dataComponentType: Pythonized JSON
        :param dataComponentType: subtype for the data array items
        :type centerComponentType: Pythonized JSON
        :param centerComponentType: subtype for the center array items
        :rtype: Pythonized JSON
        :return: a complete PFA document that performs clustering
        """

        clusters = self.pfaValue(ids, populations, sort)
        metric = self.metric.pfa("datum", "clusterCenter")
        if preprocess is None:
            preprocess = "input"

        if populations:
            populationsField = '''
    populations: int,
'''
        else:
            populationsField = ""

        return titus.prettypfa.jsonNode('''
types:
  ClusterType = record(<<clusterTypeName>>,
    id: <<idType>>,{0}
    center: array(<<centerComponentType>>))
input: array(<<dataComponentType>>)
output: <<idType>>
cells:
  clusters(array(ClusterType)) = <<clusters>>
action:
  model.cluster.closest(<<preprocess>>, clusters,
    fcn(datum: array(<<dataComponentType>>),
        clusterCenter: array(<<centerComponentType>>) -> double)
      <<metric>>)["id"]
'''.format(populationsField), **vars())
